#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""TensorBay gas CLI utility functions."""

import logging
import os
import sys
from collections import OrderedDict
from configparser import ConfigParser, SectionProxy
from functools import wraps
from typing import Any, Callable, Iterable, Optional, Sequence, Tuple, TypeVar, overload

import click
from typing_extensions import Literal, NoReturn

from tensorbay.cli.tbrn import TBRN
from tensorbay.client import GAS
from tensorbay.client import config as client_config
from tensorbay.client.dataset import DatasetClient, FusionDatasetClient
from tensorbay.client.gas import DatasetClientType
from tensorbay.client.log import dump_request_and_response
from tensorbay.client.requests import logger
from tensorbay.client.struct import Branch, Tag
from tensorbay.exception import InternalServerError, TensorBayException

_Callable = TypeVar("_Callable", bound=Callable[..., None])
_T = TypeVar("_T", Tag, Branch)
INDENT = " " * 4

_SORT_KEYS = {
    "name": lambda x: x.name,
    "commit_date": lambda x: (x.committer.date, x.name),
}


class ContextInfo:
    """This class contains command context."""

    def __init__(self, access_key: str, url: str, profile_name: str):
        self.access_key = access_key
        self.url = url
        self.profile_name = profile_name

        config_filepath = self._get_config_filepath()

        config_parser = ConfigParser(dict_type=OrderedDict)
        config_parser.read(config_filepath)
        self.config_parser = config_parser

    @staticmethod
    def _get_config_filepath() -> str:
        """Get the path of the config file.

        Returns:
            The path of the config file.

        """
        return os.path.join(os.path.expanduser("~"), ".gasconfig")

    @staticmethod
    def _is_updated(profile_section: SectionProxy) -> bool:
        return "\n" in profile_section.get("accesskey", "\n")

    @staticmethod
    def _parse_profile_info(profile_info: str) -> Tuple[str, str]:
        """Parse the profile information to get accessKey and URL.

        Arguments:
            profile_info: The profile information read from the config file.

        Returns:
            A tuple containing the accessKey and the url of profile_name read from the config file.

        """
        values = profile_info.split("\n")
        if len(values) == 2:
            return values[1], ""
        return values[1], values[2]

    def _set_request_config(self) -> None:
        """Configure request related parameters."""
        config_parser = self.config_parser
        client_config._x_source = "PYTHON-CLI"  # pylint: disable=protected-access
        if config_parser.has_section("config"):
            config_section = config_parser["config"]
            if "timeout" in config_section:
                client_config.timeout = config_section.getint("timeout")
            if "is_internal" in config_section:
                client_config.is_internal = config_section.getboolean("is_internal")
            if "max_retries" in config_section:
                client_config.max_retries = config_section.getint("max_retries")

    def generate_profiles(self) -> Iterable[Tuple[str, str, str]]:
        """Get all profiles and corresponding profile information.

        Yields:
            A tuple containing profile name and the corresponding accesskey and url read from
            the config file.

        """
        try:
            profiles = self.config_parser["profiles"]
        except KeyError:
            return
        yield from ((k, *self._parse_profile_info(v)) for k, v in profiles.items())

    def get_gas(self, access_key: Optional[str] = None, url: Optional[str] = None) -> GAS:
        """Load an object of :class:`~tensorbay.client.gas.GAS`.

        Read accessKey and URL from the appointed profile_name and login gas.

        Arguments:
            access_key: The accessKey of gas.
            url: The login URL.

        Returns:
            Gas client logged in with accessKey and URL.

        """
        self._set_request_config()

        if access_key is None:
            access_key, url = self.access_key, self.url

        if not access_key and not url:
            access_key, url = self.read_profile()

        if not access_key:
            error(
                "AccessKey should be appointed. Please visit and login to the "
                "TensorBay website to generate your AccessKey"
            )

        return GAS(access_key, url)  # type:ignore[arg-type]

    def read_profile(self) -> Tuple[str, str]:
        """Read accessKey and URL from the config file.

        Returns:
            A tuple containing the accessKey and the url of profile_name read from the config file.

        """
        profile_name = self.profile_name
        config_parser = self.config_parser
        if not config_parser.has_option("profiles", profile_name):
            error(
                f"{profile_name} does not exist"
                f"\n\nPlease use 'gas -p {profile_name} auth [accessKey]' to create profile"
            )
        return self._parse_profile_info(config_parser["profiles"][profile_name])

    def update_config(self) -> None:
        """Update the config parser to the new format."""
        config_parser = self.config_parser
        old_sections = config_parser.sections()

        if not old_sections or old_sections == ["config"]:
            return

        if (old_sections in (["config", "profiles"], ["profiles"])) and self._is_updated(
            config_parser["profiles"]
        ):
            return

        new_config_parser = ConfigParser(dict_type=OrderedDict)
        new_config_parser.add_section("profiles")
        for section_name, section_value in config_parser.items():
            if section_name == "DEFAULT":
                continue

            if section_name == "config":
                new_config_parser.add_section("config")
                new_config_parser["config"] = config_parser["config"]

            elif section_name == "profiles" and self._is_updated(section_value):
                new_config_parser["profiles"].update(section_value)

            else:
                new_config_parser["profiles"][section_name] = form_profile_value(**section_value)
        self.config_parser = new_config_parser
        self.write_config(show_message=False)

    def write_config(self, show_message: bool = True) -> None:
        """Write the config parser to the config file.

        Arguments:
            show_message: Whether to show the message.

        """
        # pylint: disable=protected-access
        config_parser = self.config_parser
        if config_parser.has_section("config"):
            config_parser._sections.move_to_end("config", last=False)  # type: ignore[attr-defined]
        if config_parser.has_section("profiles") and config_parser.has_option(
            "profiles", "default"
        ):
            config_parser._sections["profiles"].move_to_end(  # type: ignore[attr-defined]
                "default", last=False
            )

        config_file = self._get_config_filepath()
        with open(config_file, "w", encoding="utf-8") as fp:
            config_parser.write(fp)
        if show_message:
            click.echo(f'Success!\nConfiguration has been written into: "{config_file}"')


def _implement_cli(
    ctx: click.Context, access_key: str, url: str, profile_name: str, debug: bool
) -> None:
    ctx.obj = ContextInfo(access_key, url, profile_name)
    ctx.obj.update_config()

    if debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logger.disabled = True


@overload
def get_dataset_client(
    gas: GAS, tbrn_info: TBRN, is_fusion: Literal[None] = None
) -> DatasetClientType:
    ...


@overload
def get_dataset_client(gas: GAS, tbrn_info: TBRN, is_fusion: Literal[False]) -> DatasetClient:
    ...


@overload
def get_dataset_client(gas: GAS, tbrn_info: TBRN, is_fusion: Literal[True]) -> FusionDatasetClient:
    ...


@overload
def get_dataset_client(
    gas: GAS, tbrn_info: TBRN, is_fusion: Optional[bool] = None
) -> DatasetClientType:
    ...


def get_dataset_client(
    gas: GAS, tbrn_info: TBRN, is_fusion: Optional[bool] = None
) -> DatasetClientType:
    """Get the dataset client with any type and its version info.

    Arguments:
        gas: The gas client.
        tbrn_info: The tbrn of the resource.
        is_fusion: Whether the dataset is a fusion dataset, True for fusion dataset.

    Returns:
        The dataset client and its version info.

    """
    dataset_client = (
        gas._get_dataset_with_any_type(tbrn_info.dataset_name)  # pylint: disable=protected-access
        if is_fusion is None
        else gas.get_dataset(tbrn_info.dataset_name, is_fusion)
    )
    if tbrn_info.is_draft:
        dataset_client.checkout(draft_number=tbrn_info.draft_number)
    elif tbrn_info.revision is not None:
        dataset_client.checkout(revision=tbrn_info.revision)
    return dataset_client


def form_profile_value(accesskey: str, url: Optional[str] = None) -> str:
    """Form the profile value with accesskey (and url).

    Arguments:
        accesskey: The accesskey to TensorBay.
        url: The TensorBay url.

    Returns:
        The formed profile value.

    """
    values = ["", accesskey]
    if url:
        values.append(url)
    return "\n".join(values)


def _edit_input(hint: str, config_parser: ConfigParser) -> Tuple[str, str]:
    """Edit information input from the editor.

    Arguments:
        hint: The hint to be added in the temp file opened by the editor.
        config_parser: The config parser read from the context object.

    Returns:
        The extracted title and the description.

    """
    editor = config_parser["config"].get("editor") if config_parser.has_section("config") else None
    input_info = click.edit(hint, editor=editor, require_save=False)
    return _clean_up(input_info)


def _clean_up(editor_input: Optional[str]) -> Tuple[str, str]:
    """Clean up the information from the editor input.

    Arguments:
        editor_input: The editor input.

    Returns:
        The extracted title and the description.

    """
    if not editor_input:
        return "", ""

    cleaned_up_lines = []
    for line in editor_input.splitlines():
        if line and not line.startswith("#"):
            cleaned_up_lines.append(line.rstrip())

    if not cleaned_up_lines:
        return "", ""
    return cleaned_up_lines[0].lstrip(), "\n".join(cleaned_up_lines[1:])


def error(message: str) -> NoReturn:
    """Print the error message and exit the program.

    Arguments:
        message: The error message to echo.

    """
    click.secho(f"ERROR: {message}", err=True, fg="red")
    sys.exit(1)


def is_accesskey(arg: str) -> bool:
    """Determine whether the string is an AccessKey.

    Arguments:
        arg: The string needed to be judged.

    Returns:
        Whether the string is an AccessKey.

    """
    return arg.startswith(("Accesskey-", "ACCESSKEY-")) and len(arg) == 42


def shorten(origin: str) -> str:
    """Return the first 7 characters of the original string.

    Arguments:
        origin: The string needed to be shortened.

    Returns:
        A string of length 7.

    """
    return origin[:7]


def format_hint(title: str, description: str, original_hint: str) -> str:
    """Generate complete hint message.

    Arguments:
        title: The title of the draft to edit or commit.
        description: The description of the draft to edit or commit.
        original_hint: The original hint message.

    Returns:
        The complete hint message.

    """
    hint: Tuple[str, ...] = (title,)
    if description:
        hint += ("", description)
    hint += (original_hint,)
    return "\n".join(hint)


def edit_message(
    message: Tuple[str, ...], hint_message: str, config_parser: ConfigParser
) -> Tuple[str, str]:
    """Edit draft information.

    Arguments:
        message: The message given in the CLI.
        hint_message: The hint message to show on the pop-up editor.
        config_parser: The config parser read from the context object.

    Returns:
        The extracted title and the description.

    """
    if message:
        title, description = message[0], "\n".join(message[1:])
    else:
        title, description = _edit_input(hint_message, config_parser)

    return title, description


def echo_response(err: InternalServerError) -> None:
    """Log request and response when InternalServerError happens.

    Arguments:
        err: The InternalServerError raised from the CLI.

    """
    if logger.disabled:
        click.echo(dump_request_and_response(err.response))


def exception_handler(func: _Callable) -> _Callable:
    """Decorator for CLI functions to catch custom exceptions.

    Arguments:
        func: The CLI function needs to be decorated.

    Returns:
        The CLI function with exception catching procedure.

    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> None:
        try:
            func(*args, **kwargs)
        except InternalServerError as err:
            echo_response(err)
            raise
        except TensorBayException as err:
            error(str(err))

    return wrapper  # type: ignore[return-value]


def sort_branches_or_tags(sort_key: str, target: Sequence[_T]) -> Sequence[_T]:
    """Check whether the sort_key is valid and sort the target.

    Arguments:
        sort_key: The key to sort on.
        target: The target to be sorted.

    Returns:
        The sorted target.

    """
    key, reverse = (sort_key, False) if sort_key[0] != "-" else (sort_key[1:], True)
    if key not in _SORT_KEYS:
        error(
            'The "--sort" option must be "name" or "commit_date" with an optional "-" before them'
        )
    if not reverse and sort_key == "name":
        return target
    return sorted(target, key=_SORT_KEYS[key], reverse=reverse)



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of SCUT_FBP5500 dataset."""

import csv
import os
import struct
from itertools import islice

from tensorbay.dataset import Data, Dataset
from tensorbay.geometry.keypoint import Keypoint2D
from tensorbay.label import LabeledKeypoints2D
from tensorbay.utility import chunked

_DATASET_NAME = "SCUT_FBP5500"
_SEGMENT_NAMES = ("train", "test")
_CATEGORY_NAMES = {
    "AF": "Asian females",
    "AM": "Asian males",
    "CF": "Caucasian females",
    "CM": "Caucasian males",
}


def SCUT_FBP5500(path: str) -> Dataset:
    """`SCUT_FBP5500 <https://github.com/HCIILAB/SCUT-FBP5500-Database-Release>`_ dataset.

     The folder structure should be like::

        <path>
            facial landmark/
                <file_name>.pts
                ...
            Images/
                <file_name>.jpg
                ...
            train_test_files/
                split_of_60%training and 40%testing/
                    test.txt
                    train.txt
                ...
            ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    segment_path = os.path.join(
        root_path, "train_test_files", "split_of_60%training and 40%testing"
    )
    image_path = os.path.join(root_path, "Images")
    label_path = os.path.join(root_path, "facial landmark")

    dataset = Dataset(_DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for segment_name in _SEGMENT_NAMES:
        segment = dataset.create_segment(segment_name)
        with open(os.path.join(segment_path, f"{segment_name}.txt"), "r", encoding="utf-8") as fp:
            reader = csv.reader(fp, delimiter=" ")
            for filename, beauty_score in reader:
                if filename == "CM152.jpg":  # This image lacks 2D keypoints.
                    continue
                segment.append(_get_data(filename, beauty_score, image_path, label_path))
    return dataset


def _get_data(filename: str, beauty_score: str, image_path: str, label_path: str) -> Data:
    stem = os.path.splitext(os.path.basename(filename))[0]
    data = Data(os.path.join(image_path, filename))
    keypoints2d = LabeledKeypoints2D()
    keypoints2d.attributes = {"beauty_score": float(beauty_score)}
    keypoints2d.category = _CATEGORY_NAMES[stem[:2]]
    with open(os.path.join(label_path, f"{stem}.pts"), "rb") as fp:
        points = struct.unpack("i172f", fp.read())
    for x, y in chunked(islice(points, 1, None), 2):
        keypoints2d.append(Keypoint2D(float(x), float(y)))
    data.label.keypoints2d = [keypoints2d]
    return data



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of AADB dataset."""

import os
from collections import defaultdict
from typing import Callable, DefaultDict, Dict, Tuple

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification

_SEGMENTS_INFO = (
    ("new_test", "AADB_newtest", "imgListTestNewRegression"),
    ("test", "datasetImages_warp256", "imgListTestRegression"),
    ("train", "datasetImages_warp256", "imgListTrainRegression"),
    ("val", "datasetImages_warp256", "imgListValidationRegression"),
)

DATASET_NAME = "AADB"


def AADB(path: str) -> Dataset:
    """`AADB <https://www.ics.uci.edu/~skong2/aesthetics.html>`_ dataset.

    The file structure looks like:

        <path>
            AADB_newtest/
                0.500_farm1_487_20167490236_ae920475e2_b.jpg
                ...
            datasetImages_warp256/
                farm1_441_19470426814_baae1eb396_b.jpg
                ...
            imgListFiles_label/
                imgList<segment_name>Regression_<attribute_name>.txt
                ...

    Arguments:
        path: the root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    attribute_names = dataset.catalog.classification.attributes.keys()

    for mode, image_dir, label_file_prefix in _SEGMENTS_INFO:
        image_name_handler: Callable[[str, Dict[str, float]], str] = (
            (lambda image_name, attributes: f"{attributes['score']:.3f}_{image_name}")
            if mode == "new_test"
            else (lambda image_name, _: image_name)
        )

        segment = dataset.create_segment(mode)
        attributes_map = _extract_attributes_map(root_path, label_file_prefix, attribute_names)
        for image_name, attributes in attributes_map.items():
            real_image_name = image_name_handler(image_name, attributes)
            image_path = os.path.join(root_path, image_dir, real_image_name)
            data = Data(image_path)
            data.label.classification = Classification(attributes=attributes)
            segment.append(data)

    return dataset


def _extract_attributes_map(
    path: str, label_file_prefix: str, attribute_names: Tuple[str, ...]
) -> Dict[str, Dict[str, float]]:
    attributes_map: DefaultDict[str, Dict[str, float]] = defaultdict(dict)
    label_dir = os.path.join(path, "imgListFiles_label")
    for attribute_name in attribute_names:
        label_file_name = f"{label_file_prefix}_{attribute_name}.txt"
        label_file_path = os.path.join(label_dir, label_file_name)
        with open(label_file_path, encoding="utf-8") as fp:
            # one line of file looks like:
            # "<image_name> value"
            for line in fp:
                image_name, value = line.strip().split()
                attributes_map[image_name][attribute_name] = float(value)

    return attributes_map



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloaders of the 17 Category Flower dataset and the 102 Category Flower dataset."""

from tensorbay.opendataset.Flower.loader import Flower17, Flower102

__all__ = ["Flower17", "Flower102"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of HardHatWorker dataset."""

import os
from typing import Any, List

from tensorbay.dataset import Data, Dataset
from tensorbay.label import LabeledBox2D
from tensorbay.opendataset._utility import glob

try:
    import xmltodict
except ModuleNotFoundError:
    from tensorbay.opendataset._utility.mocker import xmltodict  # pylint:disable=ungrouped-imports

DATASET_NAME = "HardHatWorkers"


def HardHatWorkers(path: str) -> Dataset:
    """`Hard Hat Workers <https://makeml.app/datasets/hard-hat-workers>`_ dataset.

    The file structure should be like::

        <path>
            annotations/
                hard_hat_workers0.xml
                ...
            images/
                hard_hat_workers0.png
                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    annotation_dir = os.path.join(root_path, "annotations")

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    segment = dataset.create_segment()
    image_paths = glob(os.path.join(root_path, "images", "*.png"))
    for image_path in image_paths:
        data = Data(image_path)
        data.label.box2d = _load_labels(
            os.path.join(annotation_dir, f"{os.path.splitext(os.path.basename(image_path))[0]}.xml")
        )
        segment.append(data)
    return dataset


def _load_labels(label_file: str) -> List[LabeledBox2D]:
    with open(label_file, "r", encoding="utf-8") as fp:
        labels: Any = xmltodict.parse(fp.read())

    objects = labels["annotation"]["object"]
    box2ds = []
    if not isinstance(objects, list):
        objects = [objects]
    for obj in objects:
        bndbox = obj["bndbox"]
        box2ds.append(
            LabeledBox2D(
                xmin=float(bndbox["xmin"]),
                ymin=float(bndbox["ymin"]),
                xmax=float(bndbox["xmax"]),
                ymax=float(bndbox["ymax"]),
                category=obj["name"],
            )
        )
    return box2ds



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Data.

:class:`Data` is the most basic data unit of a :class:`~tensorbay.dataset.dataset.Dataset`.
It contains path information of a data sample and its corresponding labels.

"""

import os
from typing import Any, Dict, Optional, Type, TypeVar, Union

from tensorbay.label import Label
from tensorbay.utility import URL, FileMixin, RemoteFileMixin, ReprMixin


class DataBase(ReprMixin):
    """DataBase is a base class for the file and label combination.

    Arguments:
        timestamp: The timestamp for the file.

    Attributes:
        timestamp: The timestamp for the file.
        label: The :class:`~tensorbay.label.label.Label` instance that contains
            all the label information of the file.

    """

    _Type = Union["Data", "RemoteData", "AuthData"]

    _repr_attrs = ("timestamp", "label")

    def __init__(self, timestamp: Optional[float] = None) -> None:
        if timestamp is not None:
            self.timestamp = timestamp

        self.label = Label()


class Data(DataBase, FileMixin):
    """Data is a combination of a specific local file and its label.

    It contains the file local path, label information of the file
    and the file metadata, such as timestamp.

    A Data instance contains one or several types of labels.

    Arguments:
        local_path: The file local path.
        target_remote_path: The file remote path after uploading to tensorbay.
        timestamp: The timestamp for the file.

    Attributes:
        path: The file local path.
        timestamp: The timestamp for the file.
        label: The :class:`~tensorbay.label.label.Label` instance that contains
                all the label information of the file.
        target_remote_path: The target remote path of the data.

    """

    def __init__(
        self,
        local_path: str,
        *,
        target_remote_path: Optional[str] = None,
        timestamp: Optional[float] = None,
    ) -> None:
        DataBase.__init__(self, timestamp)
        FileMixin.__init__(self, local_path)
        self._target_remote_path = target_remote_path

    @property
    def target_remote_path(self) -> str:
        """Return the target remote path of the data.

        Target remote path will be used when this data is uploaded to tensorbay, and the target
        remote path will be the uploaded file's remote path.

        Returns:
            The target remote path of the data.

        """
        if not self._target_remote_path:
            self._target_remote_path = os.path.basename(self.path)

        return self._target_remote_path

    @target_remote_path.setter
    def target_remote_path(self, target_remote_path: str) -> None:
        self._target_remote_path = target_remote_path

    def get_callback_body(self) -> Dict[str, Any]:
        """Get the callback request body for uploading.

        Returns:
            The callback request body, which look like::

                    {
                        "remotePath": <str>,
                        "timestamp": <float>,
                        "checksum": <str>,
                        "fileSize": <int>,
                        "label": {
                            "CLASSIFICATION": {...},
                            "BOX2D": {...},
                            "BOX3D": {...},
                            "POLYGON": {...},
                            "POLYLINE2D": {...},
                            "KEYPOINTS2D": {...},
                            "SENTENCE": {...}
                        }
                    }

        """
        body = super()._get_callback_body()
        body["remotePath"] = self.target_remote_path
        if hasattr(self, "timestamp"):
            body["timestamp"] = self.timestamp
        if self.label:
            body["label"] = self.label.dumps()

        return body


class RemoteData(DataBase, RemoteFileMixin):
    """RemoteData is a combination of a specific tensorbay dataset file and its label.

    It contains the file remote path, label information of the file
    and the file metadata, such as timestamp.

    A RemoteData instance contains one or several types of labels.

    Arguments:
        remote_path: The file remote path.
        timestamp: The timestamp for the file.
        url: The URL instance used to get and update url.
        cache_path: The path to store the cache.

    Attributes:
        path: The file remote path.
        timestamp: The timestamp for the file.
        label: The :class:`~tensorbay.label.label.Label` instance that contains
                all the label information of the file.
        url: The :class:`~tensorbay.utility.file.URL` instance used to get and update url.

    """

    _T = TypeVar("_T", bound="RemoteData")

    def __init__(
        self,
        remote_path: str,
        *,
        timestamp: Optional[float] = None,
        url: Optional[URL] = None,
        cache_path: str = "",
    ) -> None:
        DataBase.__init__(self, timestamp)
        RemoteFileMixin.__init__(
            self,
            remote_path,
            url=url,
            cache_path=cache_path,
        )

    @classmethod
    def from_response_body(
        cls: Type[_T],
        body: Dict[str, Any],
        *,
        url: Optional[URL] = None,
        cache_path: str = "",
    ) -> _T:  # noqa: DAR101
        """Loads a :class:`RemoteData` object from a response body.

        Arguments:
            body: The response body which contains the information of a remote data,
                whose format should be like::

                    {
                        "remotePath": <str>,
                        "timestamp": <float>,
                        "url": <str>,
                        "label": {
                            "CLASSIFICATION": {...},
                            "BOX2D": {...},
                            "BOX3D": {...},
                            "POLYGON": {...},
                            "POLYLINE2D": {...},
                            "KEYPOINTS2D": {...},
                            "SENTENCE": {...}
                        }
                    }
            url: The URL instance used to get and update url.
            cache_path: The path to store the cache.

        Returns:
            The loaded :class:`RemoteData` object.

        """
        data = cls(
            body["remotePath"],
            timestamp=body.get("timestamp"),
            url=url,
            cache_path=cache_path,
        )
        data.label._loads(body["label"])  # pylint: disable=protected-access
        return data


class AuthData(DataBase, RemoteFileMixin):
    """AuthData is a combination of a specific cloud storaged file and its label.

    It contains the cloud storage file path, label information of the file
    and the file metadata, such as timestamp.

    An AuthData instance contains one or several types of labels.

    Arguments:
        cloud_path: The cloud file path.
        target_remote_path: The file remote path after uploading to tensorbay.
        timestamp: The timestamp for the file.
        url: The URL instance used to get and update url.

    Attributes:
        path: The cloud file path.
        timestamp: The timestamp for the file.
        label: The :class:`~tensorbay.label.label.Label` instance that contains
                all the label information of the file.
        url: The :class:`~tensorbay.utility.file.URL` instance used to get and update url.

    """

    def __init__(
        self,
        cloud_path: str,
        *,
        target_remote_path: Optional[str] = None,
        timestamp: Optional[float] = None,
        url: Optional[URL] = None,
    ) -> None:
        DataBase.__init__(self, timestamp)
        RemoteFileMixin.__init__(self, cloud_path, url=url)
        self._target_remote_path = target_remote_path

    @property
    def target_remote_path(self) -> str:
        """Return the target remote path of the auth data.

        Target remote path will be used when this data is uploaded to tensorbay, and the target
        remote path will be the uploaded file's remote path.

        Returns:
            The target remote path of the auth data.

        """
        if not self._target_remote_path:
            self._target_remote_path = self.path.split("/")[-1]

        return self._target_remote_path

    @target_remote_path.setter
    def target_remote_path(self, target_remote_path: str) -> None:
        self._target_remote_path = target_remote_path

    def get_callback_body(self) -> Dict[str, Any]:
        """Get the callback request body for uploading.

        Returns:
            The callback request body, which looks like::

                    {
                        "cloudPath": <str>,
                        "remotePath": <str>,
                        "label": {
                            "CLASSIFICATION": {...},
                            "BOX2D": {...},
                            "BOX3D": {...},
                            "POLYGON": {...},
                            "POLYLINE2D": {...},
                            "KEYPOINTS2D": {...},
                            "SENTENCE": {...}
                        }
                    }

        """
        body: Dict[str, Any] = {"cloudPath": self.path, "remotePath": self.target_remote_path}
        if self.label:
            body["label"] = self.label.dumps()
        return body



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloaders of the Kenyan Food or Nonfood dataset and Kenyan Food Type dataset."""

from tensorbay.opendataset.KenyanFood.loader import KenyanFoodOrNonfood, KenyanFoodType

__all__ = ["KenyanFoodOrNonfood", "KenyanFoodType"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Method check_basic.

:meth:`check_basic` checks whether :class:`~tensorbay.dataset.dataset.Dataset`
or :class:`~tensorbay.dataset.dataset.FusionDataset`
is empty and whether the :class:`~tensorbay.dataset.segment.Segment`
or :class:`~tensorbay.dataset.dataset.FusionDataset` in the object is empty.

"""

from typing import Iterator, Union

from tensorbay.dataset import Dataset, FusionDataset
from tensorbay.healthcheck.report import Error


class BasicError(Error):
    """The base class of the basic error.

    Arguments:
         name: The dataset or segment name which has error.

    """

    def __init__(self, name: str) -> None:
        self._name = name


class EmptyDatasetError(BasicError):
    """The health check function for empty dataset.

    This error is raised to indicate that :class:`~tensorbay.dataset.dataset.Dataset`
    or :class:`~tensorbay.dataset.dataset.FusionDataset` is empty.

    """

    def __str__(self) -> str:
        return f"Dataset '{self._name}' is empty"


class EmptySegmentError(BasicError):
    """The health check function for empty segment.

    This error is raised to indicate that :class:`~tensorbay.dataset.segment.Segment`
    or :class:`~tensorbay.dataset.dataset.FusionDataset` is empty.

    """

    def __str__(self) -> str:
        return f"Segment '{self._name}' is empty"


def check_basic(dataset: Union[Dataset, FusionDataset]) -> Iterator[BasicError]:
    """The health check function for basic error.

    Arguments:
        dataset: The :class:`~tensorbay.dataset.dataset.Dataset` or
            :class:`~tensorbay.dataset.dataset.FusionDataset` needs to be checked.

    Yields:
        BasicError indicating that :class:`~tensorbay.dataset.dataset.Dataset`,
        :class:`~tensorbay.dataset.dataset.FusionDataset`,
        :class:`~tensorbay.dataset.segment.Segment` or
        :class:`~tensorbay.dataset.segment.FusionSegment` is empty.

    """
    if not dataset:
        yield EmptyDatasetError(dataset.name)
        return

    for segment in dataset:
        if not segment:
            yield EmptySegmentError(segment.name)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the Newsgroups20 dataset."""

from tensorbay.opendataset.Newsgroups20.loader import Newsgroups20

__all__ = ["Newsgroups20"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of PASCALContext dataset."""

import os

from tensorbay.dataset import Data, Dataset
from tensorbay.exception import ModuleImportError
from tensorbay.label import SemanticMask
from tensorbay.opendataset._utility import glob

DATASET_NAME = "PASCALContext"


def PASCALContext(mask_path: str, image_path: str) -> Dataset:
    """`PASCALContext <https://cs.stanford.edu/~roozbeh/pascal-context/>`_ dataset.

    The file structure should be like::

        <mask_path>
            <image_name>.png
            ...

        <image_path>
            <image_name>.jpg
            ...

    Arguments:
        mask_path: The root directory of the dataset mask.
        image_path: The root directory of the dataset image.

    Returns:
        Loaded :class: `~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_mask_path = os.path.abspath(os.path.expanduser(mask_path))
    root_image_path = os.path.abspath(os.path.expanduser(image_path))

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    segment = dataset.create_segment("trainval")
    for mask_filename in glob(os.path.join(root_mask_path, "*.png")):
        stem = os.path.splitext(os.path.basename(mask_filename))[0]
        data = Data(os.path.join(root_image_path, f"{stem}.jpg"))
        data.label.semantic_mask = SemanticMask(mask_filename)
        segment.append(data)

    return dataset


def convert_mask(path: str, mask_path: str) -> None:
    """Convert the mat format labels of the PASCALContext dataset to masks.

    The file structure of the input path should be like::

            <path>
                <trainval>
                    <image_name>.mat
                    ...

    Arguments:
        path: The root directory of the dataset.
        mask_path: The root directory where to save the masks.

    Raises:
        ModuleImportError: When the module "scipy" or "Pillow" can not be found.

    """
    try:
        from PIL import Image  # pylint: disable=import-outside-toplevel
        from scipy.io import loadmat  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as error:
        module_name = error.name
        package_name = "Pillow" if module_name == "PIL" else None
        raise ModuleImportError(module_name=module_name, package_name=package_name) from error

    root_path = os.path.abspath(os.path.expanduser(path))
    root_mask_path = os.path.abspath(os.path.expanduser(mask_path))

    for mat_path in glob(os.path.join(root_path, "trainval", "*.mat")):
        stem = os.path.splitext(os.path.basename(mat_path))[0]
        mat = loadmat(mat_path)
        image = Image.fromarray(mat["LabelMap"])
        image.save(os.path.join(root_mask_path, f"{stem}.png"))



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of HeadPoseImage dataset."""

import os
import re
from itertools import islice
from typing import Dict, Tuple

from tensorbay.dataset import Data, Dataset
from tensorbay.label import LabeledBox2D
from tensorbay.opendataset._utility import glob

DATASET_NAME = "HeadPoseImage"


def HeadPoseImage(path: str) -> Dataset:
    """`Head Pose Image <http://crowley-coutaz.fr\
    /Head%20Pose%20Image%20Database.html>`_ dataset.

    The file structure should be like::

        <path>
            Person01/
                person01100-90+0.jpg
                person01100-90+0.txt
                person01101-60-90.jpg
                person01101-60-90.txt
                ...
            Person02/
            Person03/
            ...
            Person15/

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    segment = dataset.create_segment()
    image_paths = glob(os.path.join(path, "Person*", "*.jpg"))

    for image_path in image_paths:
        image_name = os.path.basename(image_path)
        data = Data(image_path)
        data.label.box2d = [
            LabeledBox2D(
                *_load_label_box(image_path.replace("jpg", "txt")),
                category=image_name[6:8],
                attributes=_load_attributes(image_name),
            )
        ]
        segment.append(data)
    return dataset


def _load_attributes(image_name: str) -> Dict[str, int]:
    serie = image_name[8]
    number = image_name[9:11]
    tilt, pan = re.findall(r"[\+\-]\d+", image_name)
    attributes = {
        "Serie": int(serie),
        "Number": int(number),
        "Vertical angle": int(tilt),
        "Horizontal angle": int(pan),
    }
    return attributes


def _load_label_box(label_file_path: str) -> Tuple[float, float, float, float]:
    with open(label_file_path, encoding="utf-8") as fp:
        center_x, center_y, width, height = map(int, islice(fp, 3, 7))
    return (
        center_x - width / 2,
        center_y - height / 2,
        center_x + width / 2,
        center_y + height / 2,
    )



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Label related classes."""

from tensorbay.label.attributes import AttributeInfo, Items
from tensorbay.label.catalog import Catalog
from tensorbay.label.label import Label
from tensorbay.label.label_box import Box2DSubcatalog, Box3DSubcatalog, LabeledBox2D, LabeledBox3D
from tensorbay.label.label_classification import Classification, ClassificationSubcatalog
from tensorbay.label.label_keypoints import Keypoints2DSubcatalog, LabeledKeypoints2D
from tensorbay.label.label_mask import (
    InstanceMask,
    InstanceMaskSubcatalog,
    PanopticMask,
    PanopticMaskSubcatalog,
    SemanticMask,
    SemanticMaskSubcatalog,
)
from tensorbay.label.label_polygon import (
    LabeledMultiPolygon,
    LabeledPolygon,
    LabeledRLE,
    MultiPolygonSubcatalog,
    PolygonSubcatalog,
    RLESubcatalog,
)
from tensorbay.label.label_polyline import (
    LabeledMultiPolyline2D,
    LabeledPolyline2D,
    MultiPolyline2DSubcatalog,
    Polyline2DSubcatalog,
)
from tensorbay.label.label_sentence import LabeledSentence, SentenceSubcatalog, Word
from tensorbay.label.supports import CategoryInfo, KeypointsInfo, MaskCategoryInfo

__all__ = [
    "AttributeInfo",
    "Box2DSubcatalog",
    "Box3DSubcatalog",
    "Catalog",
    "CategoryInfo",
    "Classification",
    "ClassificationSubcatalog",
    "InstanceMask",
    "InstanceMaskSubcatalog",
    "Items",
    "Keypoints2DSubcatalog",
    "KeypointsInfo",
    "Label",
    "LabeledBox2D",
    "LabeledBox3D",
    "LabeledKeypoints2D",
    "LabeledMultiPolygon",
    "LabeledMultiPolyline2D",
    "LabeledPolygon",
    "LabeledPolyline2D",
    "LabeledRLE",
    "LabeledSentence",
    "MaskCategoryInfo",
    "MultiPolygonSubcatalog",
    "MultiPolyline2DSubcatalog",
    "PanopticMask",
    "PanopticMaskSubcatalog",
    "PolygonSubcatalog",
    "Polyline2DSubcatalog",
    "RLESubcatalog",
    "SemanticMask",
    "SemanticMaskSubcatalog",
    "SentenceSubcatalog",
    "Word",
]



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of SegTrack dataset."""

import os
from typing import Callable, Dict

import numpy as np

from tensorbay.dataset import Data, Dataset
from tensorbay.label import InstanceMask
from tensorbay.opendataset._utility import glob

try:
    from PIL import Image
except ModuleNotFoundError:
    from tensorbay.opendataset._utility.mocker import Image  # pylint:disable=ungrouped-imports

DATASET_NAME = "SegTrack"

_SEGMENTS_INFO: Dict[str, Callable[[str], str]] = {
    "birdfall2": lambda stem: f"{stem}.png",
    "cheetah": lambda stem: f"chasedeer_{str(int(stem[-2:]) - 1).zfill(5)}.png",
    "girl": lambda stem: f"{int(stem[-2:]) - 61}.bmp",
    "monkeydog": lambda stem: f"Comp_00{stem}.png",
    "parachute": lambda stem: f"{stem}.png",
    "penguin": lambda stem: f"{stem}.png",
}


def SegTrack(path: str) -> Dataset:
    """`SegTrack <http://cpl.cc.gatech.edu/projects/SegTrack/>`_ dataset.

    The file structure of SegTrack looks like::

        <path>
            birdfall2/
                birdfall2_00018.png
                ...
                ground-truth/
                    birdfall2_00018.png
                    ...
            cheetah/
                chasedeer_frame_0001.bmp
                ...
                ground-truth/
                    chasedeer_00000.png
                    ...
            girl/
                5117-8_70161.bmp
                ...
                ground-truth/
                    0.bmp
                    ...
            monkeydog/
                195.bmp
                ...
                ground-truth/
                    Comp_00195.png
                    ...
            parachute/
                parachute_00000.png
                ...
                ground-truth/
                    parachute_00000.png
                    ...
            penguin/
                penguin_00000.bmp
                ...
                ground_truth/
                    penguin_00000.png
                    ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.join(os.path.abspath(os.path.expanduser(path)))
    dataset = Dataset(DATASET_NAME)
    dataset.notes.is_continuous = True
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for segment_name, filename_reformatter in _SEGMENTS_INFO.items():
        segment = dataset.create_segment(segment_name)
        segment_dir = os.path.join(root_path, segment_name)
        mask_dir = os.path.join(segment_dir, "masks")
        os.makedirs(mask_dir, exist_ok=True)
        mask_name = "ground_truth" if segment_name == "penguin" else "ground-truth"
        original_mask_dir = os.path.join(segment_dir, mask_name)
        for image_path in glob(os.path.join(segment_dir, "*.*")):
            data = Data(image_path)
            data.label.instance_mask = _get_instance_mask(
                image_path, mask_dir, original_mask_dir, filename_reformatter
            )
            segment.append(data)
    return dataset


def _get_instance_mask(
    image_path: str,
    mask_dir: str,
    original_mask_dir: str,
    filename_reformatter: Callable[[str], str],
) -> InstanceMask:
    stem = os.path.splitext(os.path.basename(image_path))[0]
    mask_path = os.path.join(mask_dir, f"{stem}.png")
    mask = np.array(
        Image.open(os.path.join(original_mask_dir, filename_reformatter(stem))),
    )[:, :, 0]
    # reformat mask
    # from {background: 0, overlap: 1~254, target: 255}
    # to {background: 0, target: 1, overlap: 255}
    overlap = np.logical_and(mask > 0, mask < 255)
    mask[mask == 255] = 1
    mask[overlap] = 255
    Image.fromarray(mask).save(mask_path)
    return InstanceMask(mask_path)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Definitions of TensorBay Resource Name (TBRN) related classes."""

from enum import Enum, Flag, auto
from typing import Any, List, Optional, Tuple, TypeVar

import click

from tensorbay.exception import TBRNError


class _TBRNFlag(Flag):
    FIELD_HEAD = 1 << 0
    FIELD_DATASET = 1 << 1
    FIELD_SEGMENT = 1 << 2
    FIELD_FRAME = 1 << 3
    FIELD_SENSOR = 1 << 4
    FIELD_PATH = 1 << 5

    # tb:[dataset]
    DATASET = FIELD_HEAD | FIELD_DATASET

    # tb:[dataset]:[segment]
    SEGMENT = FIELD_HEAD | FIELD_DATASET | FIELD_SEGMENT

    # tb:[dataset]:[segment]:[frame]
    FRAME = FIELD_HEAD | FIELD_DATASET | FIELD_SEGMENT | FIELD_FRAME

    # tb:[dataset]:[segment]::[sensor]
    SEGMENT_SENSOR = FIELD_HEAD | FIELD_DATASET | FIELD_SEGMENT | FIELD_SENSOR

    # tb:[dataset]:[segment]:[frame]:[sensor]
    FRAME_SENSOR = FIELD_HEAD | FIELD_DATASET | FIELD_SEGMENT | FIELD_FRAME | FIELD_SENSOR

    # tb:[dataset]:[segment]://[remote_path]
    NORMAL_FILE = FIELD_HEAD | FIELD_DATASET | FIELD_SEGMENT | FIELD_PATH

    # tb:[dataset]:[segment]:[frame]:[sensor]://[remote_path]
    FUSION_FILE = (
        FIELD_HEAD | FIELD_DATASET | FIELD_SEGMENT | FIELD_FRAME | FIELD_SENSOR | FIELD_PATH
    )


class TBRNType(Enum):
    """TBRNType defines the type of a TBRN.

    It has 7 types:

        1. `TBRNType.DATASET`::

            "tb:VOC2012"

        which means the dataset "VOC2012".

        2. `TBRNType.SEGMENT`::

            "tb:VOC2010:train"

        which means the "train" segment of dataset "VOC2012".

        3. `TBRNType.FRAME`::

            "tb:KITTI:test:10"

        which means the 10th frame of the "test" segment in dataset "KITTI".

        4. `TBRNType.SEGMENT_SENSOR`::

            "tb:KITTI:test::lidar"

        which means the sensor "lidar" of the "test" segment in dataset "KITTI".

        5. `TBRNType.FRAME_SENSOR`::

            "tb:KITTI:test:10:lidar"

        which means the sensor "lidar" which belongs to the 10th frame of the "test" segment in
        dataset "KITTI".

        6. `TBRNType.NORMAL_FILE`::

            "tb:VOC2012:train://2012_004330.jpg"

        which means the file "2012_004330.jpg" of the "train" segment in normal dataset "VOC2012".

        7. `TBRNType.FUSION_FILE`::

            "tb:KITTI:test:10:lidar://000024.bin"

        which means the file "000024.bin" in fusion dataset "KITTI", its segment, frame index and
        sensor is "test", 10 and "lidar".

    """

    DATASET = auto()
    SEGMENT = auto()
    FRAME = auto()
    SEGMENT_SENSOR = auto()
    FRAME_SENSOR = auto()
    NORMAL_FILE = auto()
    FUSION_FILE = auto()


_T = TypeVar("_T")


class TBRN:
    """TBRN is a TensorBay Resource Name(TBRN) parser and generator.

    Use as a generator:

        >>> tbrn_info = TBRN("VOC2010", "train", remote_path="2012_004330.jpg")
        >>> tbrn_info.type
        <TBRNType.NORMAL_FILE: 5>
        >>> tbrn_info.get_tbrn()
        'tb:VOC2010:train://2012_004330.jpg'
        >>> print(tbrn_info)
        'tb:VOC2010:train://2012_004330.jpg'

    Use as a parser:

        >>> tbrn = "tb:VOC2010:train://2012_004330.jpg"
        >>> tbrn_info = TBRN(tbrn=tbrn)
        >>> tbrn_info.dataset
        'VOC2010'
        >>> tbrn_info.segment_name
        'train'
        >>> tbrn_info.remote_path
        '2012_004330.jpg'

    Arguments:
        dataset_name: Name of the dataset.
        segment_name: Name of the segment.
        frame_index: Index of the frame.
        sensor_name: Name of the sensor.
        remote_path: Object path of the file.
        tbrn: Full TBRN string.
        draft_number: The draft number (if the status is draft).
        revision: The commit revision (if the status is commit).

    Attributes:
        dataset_name: Name of the dataset.
        segment_name: Name of the segment.
        frame_index: Index of the frame.
        sensor_name: Name of the sensor.
        remote_path: Object path of the file.
        type: The type of this TBRN.
        draft_number: The draft number (if the status is draft).
        revision: The revision (if the status is not draft).
        is_draft: whether the status is draft, True for draft, False for commit.

    Raises:
        TBRNError: The TBRN is invalid.

    """

    _HEAD = "tb"
    _NAMES_SEPARATOR = ":"

    _DRAFT_SEPARATOR = "#"
    _REVISION_SEPARATOR = "@"

    _NAMES_MAX_SPLIT = 4
    _PATH_SEPARATOR = "://"
    _PATH_MAX_SPLIT = 1
    _FRAME_INDEX = 3

    _FLAG_TO_TYPE = {
        _TBRNFlag.DATASET: (TBRNType.DATASET, 1),
        _TBRNFlag.SEGMENT: (TBRNType.SEGMENT, 2),
        # _TBRNFlag.FRAME: (TBRNType.FRAME, 3),
        # _TBRNFlag.SEGMENT_SENSOR: (TBRNType.SEGMENT_SENSOR, 4),
        # _TBRNFlag.FRAME_SENSOR: (TBRNType.FRAME_SENSOR, 4),
        _TBRNFlag.NORMAL_FILE: (TBRNType.NORMAL_FILE, 2),
        # _TBRNFlag.FUSION_FILE: (TBRNType.FUSION_FILE, 4),
    }

    _names: Tuple[
        Optional[str],
        Optional[str],
        Optional[int],
        Optional[str],
        Optional[str],
    ]

    def __init__(
        self,
        dataset_name: Optional[str] = None,
        segment_name: Optional[str] = None,
        frame_index: Optional[int] = None,
        sensor_name: Optional[str] = None,
        *,
        remote_path: Optional[str] = None,
        draft_number: Optional[int] = None,
        revision: Optional[str] = None,
        tbrn: Optional[str] = None,
    ) -> None:
        if tbrn is not None:
            splits = tbrn.split(TBRN._PATH_SEPARATOR, TBRN._PATH_MAX_SPLIT)
            names: List[Any] = splits[0].split(TBRN._NAMES_SEPARATOR, TBRN._NAMES_MAX_SPLIT)

            if names[0] != TBRN._HEAD:
                raise TBRNError('TensorBay Resource Name should startwith "tb:"')

            dataset_name = names[1]
            if not dataset_name:
                raise TBRNError(
                    'TensorBay Resource Name should add dataset name "tb:<dataset name>"'
                )

            self.revision: Optional[str] = None
            self.draft_number: Optional[int] = None

            if TBRN._REVISION_SEPARATOR in dataset_name:
                names[1], self.revision = dataset_name.split(TBRN._REVISION_SEPARATOR, 1)

            elif TBRN._DRAFT_SEPARATOR in dataset_name:
                names[1], number = dataset_name.split(TBRN._DRAFT_SEPARATOR)
                self.draft_number = int(number)
            else:
                names[1] = dataset_name

            names += [None] * (TBRN._NAMES_MAX_SPLIT + 1 - len(names))

            frame_index = names[TBRN._FRAME_INDEX]
            names[TBRN._FRAME_INDEX] = int(frame_index) if frame_index else None

            if len(splits) == TBRN._PATH_MAX_SPLIT + 1:
                names.append(splits[1])
            else:
                names.append(None)

            self._names = tuple(names[1:])  # type: ignore[assignment]

        else:
            self._names = (
                dataset_name,
                segment_name,
                frame_index,
                sensor_name,
                remote_path,
            )

            if draft_number is not None and revision is not None:
                raise TBRNError("TensorBay Resource Name should not contain draft and commit info")

            self.draft_number = draft_number
            self.revision = revision

        try:
            self._type, self._field_length = self._check_type()
        except KeyError as error:
            raise TBRNError("Invalid TensorBay Resource Name") from error

    def __repr__(self) -> str:
        return self.get_tbrn()

    def _check_type(self) -> Tuple[TBRNType, int]:
        flag = _TBRNFlag.FIELD_HEAD.value
        for i, name in enumerate(self._names, 1):
            if name is not None:
                flag |= 1 << i

        return TBRN._FLAG_TO_TYPE[_TBRNFlag(flag)]

    def _raise_when_none(self, value: Optional[_T], field: str) -> _T:
        if value is None:
            raise ValueError(f"TBRNType: '{self.type.name}' has no property '{field}'")
        return value

    @property
    def dataset_name(self) -> str:
        """Return the dataset name.

        Returns:
            The dataset name.

        """
        return self._names[0]  # type: ignore[return-value]

    @property
    def segment_name(self) -> str:
        """Return the segment name.

        Returns:
            The segment name.

        """
        return self._raise_when_none(self._names[1], "segment_name")

    @property
    def frame_index(self) -> int:
        """Return the frame index.

        Returns:
            The frame index.

        """
        return self._raise_when_none(self._names[2], "frame_index")

    @property
    def sensor_name(self) -> str:
        """Return the sensor name.

        Returns:
            The sensor name.

        """
        return self._raise_when_none(self._names[3], "sensor_name")

    @property
    def remote_path(self) -> str:
        """Return the object path.

        Returns:
            The object path.

        """
        return self._raise_when_none(self._names[4], "remote_path")

    @property
    def type(self) -> TBRNType:
        """Return the type of this TBRN.

        Returns:
            The type of this TBRN.

        """
        return self._type

    @property
    def is_draft(self) -> bool:
        """Return the frame index.

        Returns:
            The frame index.

        """
        return bool(self.draft_number)

    def get_tbrn(self, frame_width: int = 0) -> str:
        """Generate the full TBRN string.

        Arguments:
            frame_width: Add '0' at the beginning of the frame_index,
                until it reaches the frame_width.

        Returns:
            The full TBRN string.

        """
        names: List[Any] = [TBRN._HEAD]
        names.extend(self._names[: self._field_length])
        if self._field_length >= TBRN._FRAME_INDEX:
            frame_index = names[TBRN._FRAME_INDEX]
            if frame_index is not None:
                if frame_width:
                    names[TBRN._FRAME_INDEX] = f"{frame_index:0{frame_width}}"
                else:
                    names[TBRN._FRAME_INDEX] = f"{frame_index}"
            else:
                names[TBRN._FRAME_INDEX] = ""
        if self.is_draft:
            names[1] = f"{names[1]}{TBRN._DRAFT_SEPARATOR}{self.draft_number}"
        elif self.revision is not None:
            names[1] = f"{names[1]}{TBRN._REVISION_SEPARATOR}{self.revision}"
        tbrn = TBRN._NAMES_SEPARATOR.join(names)
        if self._names[4] is not None:
            tbrn = f"{tbrn}{TBRN._PATH_SEPARATOR}{self.remote_path}"
        return tbrn

    def get_colored_tbrn(self, color: Optional[str] = "yellow") -> str:
        """Get the colored tbrn string.

        Arguments:
            color: The color of the tbrn.

        Returns:
            The colored tbrn string.

        """
        return click.style(self.get_tbrn(), fg=color)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""AttrsMixin and Field class.

:class:`AttrsMixin` provides a list of special methods based on field configs.

:class:`Field` is a class describing the attr related fields.

"""
from sys import version_info
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    no_type_check,
)

from typing_extensions import NoReturn, Protocol

from tensorbay.exception import AttrError

_T = TypeVar("_T")
_Callable = Callable[[Any], Any]
_KeyConverter = Callable[[str], str]
_BUILTINS = {"builtins", None, "typing"}
_DEFAULT_ERROR_MESSAGE = "'{class_name}' object has no attribute '{attr_name}'"
_ATTRS_BASE = "_attrs_base"


class _A(Protocol):
    def dumps(self) -> Any:
        """Dumps all the information of attrs into a dict."""


class Field:
    """A class to identify attr fields.

    Arguments:
        is_dynamic: Whether attr is a dynamic attr.
        key: Display value of the attr in contents.
        default: Default value of the attr.
        error_message: The custom error message of the attr.
        loader: The custom loader of the attr.
        dumper: The custom dumper of the attr.

    """

    # remove __slots__ and add __getattr__ due to: https://github.com/PyCQA/pylint/issues/4341
    # __slots__ = ("is_dynamic", "key", "default", "error_message", "loader", "dumper")

    def __init__(
        self,
        *,
        is_dynamic: bool,
        key: Union[str, None, _KeyConverter],
        default: Any,
        error_message: Optional[str],
        loader: Optional[_Callable],
        dumper: Optional[_Callable],
    ) -> None:
        if loader:
            self.loader = loader

        if dumper:
            self.dumper = dumper

        self.is_dynamic = is_dynamic
        self.default = default

        if callable(key):
            self.key_converter = key
        else:
            self.key = key

        if error_message:
            self.error_message = error_message

    def __getattr__(self, name: str) -> NoReturn:
        raise AttributeError(
            _DEFAULT_ERROR_MESSAGE.format(class_name=self.__class__.__name__, attr_name=name)
        )


class BaseField:
    """A class to identify fields of base class.

    Arguments:
        key: Display value of the attr.

    """

    def __init__(self, key: Optional[str]) -> None:
        self.loader: _Callable
        self.dumper: _Callable
        self.key = key


class AttrsMixin:
    """AttrsMixin provides a list of special methods based on attr fields.

    Examples:
        box2d: Box2DSubcatalog = attr(is_dynamic=True, key="BOX2D")

    """

    _attrs_fields: Dict[str, Field]
    _attrs_base: Any

    def __init_subclass__(cls) -> None:
        type_ = cls.__annotations__.pop(_ATTRS_BASE, None)
        if type_:
            cls._attrs_base.loader = type_._loads  # pylint: disable=protected-access
            cls._attrs_base.dumper = getattr(type_, "_dumps", type_.dumps)

        attrs_fields = {}
        for base_class in cls.__bases__:
            base_fields = getattr(base_class, "_attrs_fields", None)
            if base_fields:
                attrs_fields.update(base_fields)
        for name, type_ in getattr(cls, "__annotations__", {}).items():
            field = getattr(cls, name, None)
            if isinstance(field, Field):
                need_loader = not hasattr(field, "loader")
                need_dumper = not hasattr(field, "dumper")
                if need_loader or need_dumper:
                    loader, dumper = _get_operators(type_)
                    if need_loader:
                        field.loader = loader
                    if need_dumper:
                        field.dumper = dumper

                if hasattr(field, "key_converter"):
                    field.key = field.key_converter(name)
                attrs_fields[name] = field
                delattr(cls, name)
        cls._attrs_fields = attrs_fields

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return self.__dict__ == other.__dict__

    @no_type_check
    def __getattr__(self, name: str) -> NoReturn:
        """Raise an AttributeError exception when an attr does not exist.

        Arguments:
            name: Name of the attr.

        Raises:
            AttributeError: When an attr does not exist.

        """
        error_message = getattr(
            self._attrs_fields.get(name), "error_message", _DEFAULT_ERROR_MESSAGE
        )
        raise AttributeError(
            error_message.format(class_name=self.__class__.__name__, attr_name=name)
        )

    def _loads(self, contents: Any) -> None:
        """Load an instance from a dict containing the information of the contents.

        Arguments:
            contents: A dict containing all data.

        """
        base = getattr(self, _ATTRS_BASE, None)
        if base:
            value = contents if base.key is None else contents[base.key]
            base.loader(self, value)

        for name, field in self._attrs_fields.items():
            if field.is_dynamic and field.key not in contents:
                continue

            if field.key is None:
                value = contents
            elif field.default is not ...:
                value = contents.get(field.key, field.default)
            else:
                value = contents[field.key]

            setattr(self, name, field.loader(value))

    def _dumps(self) -> Dict[str, Any]:
        """Dumps all the information of attrs into a dict.

        Returns:
            contents: A dict containing all data of attrs.

        """
        contents: Dict[str, Any] = {}
        base = getattr(self, _ATTRS_BASE, None)
        if base:
            _key_dumper(base.key, contents, base.dumper(self))

        for name, field in self._attrs_fields.items():
            if field.is_dynamic and not hasattr(self, name):
                continue

            value = getattr(self, name)
            if value == field.default:
                continue

            _key_dumper(field.key, contents, field.dumper(value))
        return contents


def attr(
    *,
    is_dynamic: bool = False,
    key: Union[str, None, _KeyConverter] = lambda x: x,
    default: Any = ...,
    error_message: Optional[str] = None,
    loader: Optional[Callable[[Any], Any]] = None,
    dumper: Optional[Callable[[Any], Any]] = None,
) -> Any:
    """Return an instance to identify attr fields.

    Arguments:
        is_dynamic: Determine if this is a dynamic attr.
        key: Display value of the attr in contents.
        default: Default value of the attr.
        error_message: The custom error message of the attr.
        loader: The custom loader of the attr.
        dumper: The custom dumper of the attr.

    Raises:
        AttrError: Dynamic attr cannot have default value.

    Returns:
        A :class:`Field` instance containing all attr fields.

    """
    if is_dynamic and default is not ...:
        raise AttrError()

    return Field(
        is_dynamic=is_dynamic,
        key=key,
        default=default,
        error_message=error_message,
        loader=loader,
        dumper=dumper,
    )


def attr_base(key: Optional[str] = None) -> Any:
    """Return an instance to identify base class fields.

    Arguments:
        key: Display value of the attr.

    Returns:
        A :class:`BaseField` instance containing all base class fields.

    """
    return BaseField(key)


def upper(name: str) -> str:
    """Convert the name value to uppercase.

    Arguments:
        name: name of the attr.

    Returns:
        The uppercase value.
    """
    return name.upper()


def camel(name: str) -> str:
    """Convert the name value to camelcase.

    Arguments:
        name: name of the attr.

    Returns:
        The camelcase value.

    """
    mixed = name.title().replace("_", "")
    return f"{mixed[0].lower()}{mixed[1:]}"


def _key_dumper(key: Optional[str], contents: Dict[str, Any], value: Any) -> None:
    if key is None:
        contents.update(value)
    else:
        contents[key] = value


def _get_origin_in_3_7(annotation: Any) -> Any:
    return getattr(annotation, "__origin__", None)


def _get_origin_in_3_6(annotation: Any) -> Any:
    module = getattr(annotation, "__module__", None)
    if module in _BUILTINS:
        extra = getattr(annotation, "__extra__", None)
        if extra is not None:
            return extra

    return getattr(annotation, "__origin__", None)


_get_origin = _get_origin_in_3_6 if version_info < (3, 7) else _get_origin_in_3_7


def _get_operators(annotation: Any) -> Tuple[_Callable, _Callable]:
    """Get attr operating methods by annotations.

     AttrsMixin has three operating types which are classified by attr annotation.
        1. builtin types, like str, int, None
        2. tensorbay custom class, like tensorbay.label.Classification
        3. tensorbay custom class list or NameList, like List[tensorbay.label.LabeledBox2D]

    Arguments:
        annotation: Type of the attr.

    Returns:
        Operating methods of the annotation.

    """
    origin: Any = _get_origin(annotation)
    annotation_args = getattr(annotation, "__args__", ())
    if isinstance(origin, type) and issubclass(origin, Sequence):
        type_ = annotation_args[0]
    elif origin is Union and len(annotation_args) == 2 and type(None) in annotation_args:
        type_ = annotation_args[0] if isinstance(None, annotation_args[1]) else annotation_args[1]
    else:
        origin = None
        type_ = annotation

    if {getattr(origin, "__module__", None), getattr(type_, "__module__", None)} < _BUILTINS:
        return _builtin_operator, _builtin_operator

    if origin is None:
        return type_.loads, _attr_dumper

    if origin is Union:
        return (
            lambda contents: type_.loads(contents) if contents is not None else None,
            _attr_optional_dumper,
        )

    return (
        lambda contents: origin(type_.loads(content) for content in contents),
        _attr_list_dumper,
    )


def _builtin_operator(contents: _T) -> _T:
    return contents


def _attr_dumper(attr_: _A) -> Any:
    return attr_.dumps()


def _attr_optional_dumper(attr_: Optional[_A]) -> Any:
    if attr_ is not None:
        return attr_.dumps()
    return None


def _attr_list_dumper(attrs: List[_A]) -> List[Any]:
    return [attr_.dumps() for attr_ in attrs]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Frame.

:class:`Frame` is a concept in :class:`~tensorbay.dataset.dataset.FusionDataset`.

It is the structure that composes a :class:`~tensorbay.dataset.segment.FusionSegment`,
and consists of multiple :class:`~tensorbay.dataset.data.Data` collected at the same time
from different sensors.

"""

import logging
from typing import Any, Dict, Optional, Type, TypeVar
from uuid import UUID

from ulid import ULID, from_str, from_uuid

from tensorbay.client.lazy import LazyPage
from tensorbay.dataset.data import DataBase, RemoteData
from tensorbay.utility import URL, UserMutableMapping

logger = logging.getLogger(__name__)


class Frame(UserMutableMapping[str, "DataBase._Type"]):
    """This class defines the concept of frame.

    Frame is a concept in :class:`~tensorbay.dataset.dataset.FusionDataset`.

    It is the structure that composes :class:`~tensorbay.dataset.segment.FusionSegment`,
    and consists of multiple :class:`~tensorbay.dataset.data.Data` collected at the same time
    corresponding to different sensors.

    Since :class:`Frame` extends :class:`~tensorbay.utility.user.UserMutableMapping`,
    its basic operations are the same as a dictionary's.

    To initialize a Frame and add a :class:`~tensorbay.dataset.data.Data` to it:

    .. code:: python

        frame = Frame()
        frame[sensor_name] = Data()

    """

    _T = TypeVar("_T", bound="Frame")
    _logger_flag = True

    def __init__(self, frame_id: Optional[ULID] = None) -> None:
        self._data: Dict[str, DataBase._Type] = {}
        # self._pose: Optional[Transform3D] = None
        if frame_id:
            self.frame_id = frame_id

    def _repr_head(self) -> str:
        if hasattr(self, "frame_id"):
            return f'{self.__class__.__name__}("{self.frame_id}")'

        return self.__class__.__name__

    @classmethod
    def from_response_body(
        cls: Type[_T],
        body: Dict[str, Any],
        url_index: int,
        urls: LazyPage[Dict[str, str]],
        *,
        cache_path: str = "",
    ) -> _T:
        """Loads a :class:`Frame` object from a response body.

        Arguments:
            body: The response body which contains the information of a frame,
                whose format should be like::

                    {
                        "frameId": <str>,
                        "frame": [
                            {
                                "sensorName": <str>,
                                "remotePath": <str>,
                                "timestamp": <float>,
                                "url": <str>,
                                "label": {...}
                            },
                            ...
                            ...
                        ]
                    }
            url_index: The index of the url.
            urls: A sequence of mappings which key is the sensor name and value is the url.
            cache_path: The path to store the cache.

        Returns:
            The loaded :class:`Frame` object.

        """  # noqa: DAR101  # https://github.com/terrencepreilly/darglint/issues/120
        try:
            frame_id = from_str(body["frameId"])
        except ValueError:
            # Legacy fusion dataset use uuid as frame ID
            # Keep this code here to make SDK compatible with uuid
            frame_id = from_uuid(UUID(body["frameId"]))
            if cls._logger_flag:
                cls._logger_flag = False
                logger.warning(
                    "WARNING: This is a legacy fusion dataset which use uuid as frame ID, "
                    "it should be updated to ulid."
                )
        frame = cls(frame_id)
        for data_contents in body["frame"]:
            sensor_name = data_contents["sensorName"]
            url = URL.from_getter(lambda s=sensor_name: urls.items[url_index].get()[s], urls.pull)
            frame[sensor_name] = RemoteData.from_response_body(
                data_contents,
                url=url,
                cache_path=cache_path,
            )
        return frame

    # @property
    # def pose(self) -> Optional[Transform3D]:
    #     """Return the pose of the frame.

    #     Returns:
    #         A :class:`~tensorbay.geometry.transform.Transform3D` object
    #         representing the pose of the frame.

    #     """
    #     return self._pose

    # def set_pose(
    #     self,
    #     translation: Optional[Iterable[float]] = (0, 0, 0),
    #     rotation: Transform3D.RotationType = (1, 0, 0, 0),
    #     *,
    #     matrix: Optional[MatrixType] = None,
    # ) -> None:
    #     """Set the pose of the current frame.

    #     Arguments:
    #         translation: Translation of the frame pose in a sequence of [x, y, z].
    #         rotation: Rotation of the frame pose in a sequence of [w, x, y, z]
    #             or a numpy quaternion object.
    #         matrix: The transform representing the frame pose in
    #             a 4x4 or 3x4 transform matrix.

    #     """
    #     self._pose = Transform3D(translation, rotation, matrix=matrix)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of NightOwls dataset."""

import json
import os
from typing import Any, Dict, Optional

from tensorbay.dataset import Data, Dataset
from tensorbay.label import LabeledBox2D
from tensorbay.opendataset._utility import coco, glob

DATASET_NAME = "NightOwls"


def NightOwls(path: str) -> Dataset:
    """`NightOwls <http://www.nightowls-dataset.org/>`_ dataset.

    The file structure should be like::

        <path>
            nightowls_test/
                <image_name>.png
                ...
            nightowls_training/
                <image_name>.png
                ...
            nightowls_validation/
                <image_name>.png
                ...
            nightowls_training.json
            nightowls_validation.json

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME)
    dataset.notes.is_continuous = True
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for mode, (labels_filename, labels_handler) in _LABELS_HANDEL_METHODS.items():
        segment = dataset.create_segment(mode)

        image_paths = glob(os.path.join(root_path, f"nightowls_{mode}", "*.png"))

        labels = _load_labels(root_path, labels_filename)

        for image_path in image_paths:
            data = labels_handler(image_path, labels)  # pylint: disable=not-callable
            segment.append(data)

    return dataset


def _load_labels(root_path: str, labels_filename: Optional[str]) -> Dict[str, Any]:
    if not labels_filename:
        return {"test": None}

    label_path = os.path.join(root_path, labels_filename)
    with open(label_path, "r", encoding="utf-8") as fp:
        loaded_data = json.load(fp)

    labels = {}

    image_name_id_map = {image["file_name"]: image["id"] for image in loaded_data["images"]}
    labels["image_name_id_map"] = image_name_id_map
    labels["poses"] = loaded_data["poses"]

    # the origin name of forth pose is "nan", this expression changes it to None
    labels["poses"][4]["name"] = None

    coco_labels = coco(os.path.join(root_path, labels_filename))

    labels["images"] = coco_labels.images
    labels["annotations"] = coco_labels.annotations
    labels["image_annotations_map"] = coco_labels.image_annotations_map
    labels["categories"] = coco_labels.categories

    return labels


def _generate_data(image_path: str, labels: Dict[str, Any]) -> Data:
    data = Data(image_path)
    data.label.box2d = []

    image_id = labels["image_name_id_map"][os.path.basename(image_path)]
    image_annotations_map = labels["image_annotations_map"]

    if image_id not in image_annotations_map:
        return data

    annotations = labels["annotations"]
    poses = labels["poses"]
    categories = labels["categories"]

    for annotation_id in image_annotations_map[image_id]:
        annotation = annotations[annotation_id]
        x_top, y_top, width, height = annotation["bbox"]

        attributes = {
            "occluded": annotation["occluded"],
            "difficult": annotation["difficult"],
            "pose": poses[annotation["pose_id"] - 1]["name"],
            "truncated": annotation["truncated"],
        }

        data.label.box2d.append(
            LabeledBox2D.from_xywh(
                x=x_top,
                y=y_top,
                width=width,
                height=height,
                category=categories[annotation["category_id"]]["name"],
                attributes=attributes,
                instance=str(annotation["tracking_id"]),
            )
        )

    return data


_LABELS_HANDEL_METHODS = {
    "training": ("nightowls_training.json", _generate_data),
    "validation": ("nightowls_validation.json", _generate_data),
    "test": (None, lambda image_path, _: Data(image_path)),
}



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Notes, DatasetBase, Dataset and FusionDataset.

:class:`Notes` contains the basic information of a :class:`DatasetBase`.

:class:`DatasetBase` defines the basic concept of a dataset,
which is the top-level structure to handle your data files, labels and other additional information.

It represents a whole dataset contains several segments
and is the base class of :class:`Dataset` and :class:`FusionDataset`.

:class:`Dataset` is made up of data collected from only one sensor
or data without sensor information.
It consists of a list of :class:`~tensorbay.dataset.segment.Segment`.

:class:`FusionDataset` is made up of data collected from multiple sensors.
It consists of a list of :class:`~tensorbay.dataset.segment.FusionSegment`.

"""

import json
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    KeysView,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from tensorbay.dataset.segment import FusionSegment, Segment
from tensorbay.label import Catalog
from tensorbay.utility import (
    AttrsMixin,
    NameMixin,
    ReprMixin,
    ReprType,
    SortedNameList,
    attr,
    camel,
    common_loads,
    locked,
)

if TYPE_CHECKING:
    from tensorbay.client import GAS

_T = TypeVar("_T", FusionSegment, Segment)


class Notes(AttrsMixin, ReprMixin):
    """This is a class stores the basic information of :class:`DatasetBase`.

    Arguments:
        is_continuous: Whether the data inside the dataset is time-continuous.
        bin_point_cloud_fields: The field names of the bin point cloud files in the dataset.

    """

    _T = TypeVar("_T", bound="Notes")

    _repr_attrs = ("is_continuous", "bin_point_cloud_fields")

    is_continuous: bool = attr(key=camel, default=False)
    bin_point_cloud_fields: Optional[List[str]] = attr(key=camel, default=None)

    def __init__(
        self, is_continuous: bool = False, bin_point_cloud_fields: Optional[Iterable[str]] = None
    ) -> None:
        self.is_continuous = is_continuous
        self.bin_point_cloud_fields = (
            list(bin_point_cloud_fields) if bin_point_cloud_fields else None
        )

    def __getitem__(self, key: str) -> Any:
        try:
            return getattr(self, key)
        except AttributeError as error:
            raise KeyError(key) from error

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Loads a :class:`Notes` instance from the given contents.

        Arguments:
            contents: The given dict containing the dataset notes::

                    {
                        "isContinuous":            <boolean>
                        "binPointCloudFields": [   <array> or null
                                <field_name>,      <str>
                                ...
                        ]
                    }

        Returns:
            The loaded :class:`Notes` instance.

        """
        return common_loads(cls, contents)

    def keys(self) -> KeysView[str]:
        """Return the valid keys within the notes.

        Returns:
            The valid keys within the notes.

        """
        return KeysView(self._repr_attrs)  # type: ignore[arg-type]

    def dumps(self) -> Dict[str, Any]:
        """Dumps the notes into a dict.

        Returns:
            A dict containing all the information of the Notes::

                {
                    "isContinuous":           <boolean>
                    "binPointCloudFields": [  <array> or null
                        <field_name>,         <str>
                        ...
                    ]
                }

        """
        return self._dumps()


# When the NameMixin is before Sequence[_T], typing will raise AttributeError.
# related issue: python/typing#777
class DatasetBase(Sequence[_T], NameMixin):
    """This class defines the concept of a basic dataset.

    DatasetBase represents a whole dataset contains several segments
    and is the base class of :class:`Dataset` and :class:`FusionDataset`.

    A dataset with labels should contain a :class:`~tensorbay.label.catalog.Catalog`
    indicating all the possible values of the labels.

    Arguments:
        name: The name of the dataset.
        gas: The :class:`~tensorbay.client.gas.GAS` client for getting a remote dataset.
        revision: The revision of the remote dataset.

    Attributes:
        catalog: The :class:`~tensorbay.label.catalog.Catalog` of the dataset.
        notes: The :class:`Notes` of the dataset.

    """

    _is_fusion: bool

    _repr_type = ReprType.SEQUENCE

    def __init__(
        self,
        name: str,
        gas: Optional["GAS"] = None,
        revision: Optional[str] = None,
    ) -> None:
        super().__init__(name)

        if gas:
            self._client = gas.get_dataset(name, is_fusion=self._is_fusion)
            if revision:
                self._client.checkout(revision)
        else:
            self._segments: SortedNameList[_T] = SortedNameList()
            self._catalog = Catalog()
            self._notes = Notes()

    def __len__(self) -> int:
        return self._get_segments().__len__()

    @overload
    def __getitem__(self, index: Union[int, str]) -> _T:
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[_T]:
        ...

    def __getitem__(self, index: Union[int, str, slice]) -> Union[Sequence[_T], _T]:
        return self._get_segments().__getitem__(index)

    def __delitem__(self, index: Union[int, str, slice]) -> None:
        return self._get_segments().__delitem__(index)

    def __contains__(self, key: Any) -> bool:
        return self._get_segments().__contains__(key)

    def _assert_remote(self) -> None:
        if not hasattr(self, "_client"):
            raise TypeError("Cache is only for remote datasets")

    @locked
    def _init_segments(self) -> None:
        self._segments = SortedNameList()

        for segment in self._client._list_segment_instances():  # pylint: disable=protected-access
            self._segments.add(segment)  # type: ignore[arg-type]

    def _get_segments(self) -> SortedNameList[_T]:
        if not hasattr(self, "_segments"):
            self._init_segments()

        return self._segments

    @property
    def catalog(self) -> Catalog:
        """Return the catalog of the dataset.

        Returns:
            The :class:`~tensorbay.label.catalog.Catalog` of the dataset.

        """
        if not hasattr(self, "_catalog"):
            self._catalog = self._client.get_catalog()

        return self._catalog

    @property
    def notes(self) -> Notes:
        """Return the notes of the dataset.

        Returns:
            The class:`Notes` of the dataset.

        """
        if not hasattr(self, "_notes"):
            self._notes = self._client.get_notes()

        return self._notes

    @property
    def cache_enabled(self) -> bool:
        """Whether the cache is enabled.

        Returns:
            Whether the cache is enabled.

        """
        self._assert_remote()
        return self._client.cache_enabled

    def enable_cache(self, cache_path: str = "") -> None:
        """Enable cache when open the remote data of the dataset.

        Arguments:
            cache_path: The path to store the cache.

        """
        self._assert_remote()
        self._client.enable_cache(cache_path)

    def keys(self) -> Tuple[str, ...]:
        """Get all segment names.

        Returns:
            A tuple containing all segment names.

        """
        return self._get_segments().keys()

    def load_catalog(self, filepath: str) -> None:
        """Load catalog from a json file.

        Arguments:
            filepath: The path of the json file which contains the catalog information.

        """
        with open(filepath, "r", encoding="utf-8") as fp:
            contents = json.load(fp)
        self._catalog = Catalog.loads(contents)

    def add_segment(self, segment: _T) -> None:
        """Add a segment to the dataset.

        Arguments:
            segment: The segment to be added.

        """
        self._get_segments().add(segment)


class Dataset(DatasetBase[Segment]):
    """This class defines the concept of dataset.

    Dataset is made up of data collected from only one sensor or data without sensor information.
    It consists of a list of :class:`~tensorbay.dataset.segment.Segment`.

    """

    _is_fusion = False

    def create_segment(self, segment_name: str = "default") -> Segment:
        """Create a segment with the given name.

        Arguments:
            segment_name: The name of the segment to create, which default value is an empty string.

        Returns:
            The created :class:`~tensorbay.dataset.segment.Segment`.

        """
        segment = Segment(segment_name)
        self._get_segments().add(segment)
        return segment


class FusionDataset(DatasetBase[FusionSegment]):
    """This class defines the concept of fusion dataset.

    FusionDataset is made up of data collected from multiple sensors.
    It consists of a list of :class:`~tensorbay.dataset.segment.FusionSegment`.
    """

    _is_fusion = True

    def create_segment(self, segment_name: str = "default") -> FusionSegment:
        """Create a fusion segment with the given name.

        Arguments:
            segment_name: The name of the fusion segment to create,
                which default value is an empty string.

        Returns:
            The created :class:`~tensorbay.dataset.segment.FusionSegment`.

        """
        segment = FusionSegment(segment_name)
        self._get_segments().add(segment)
        return segment



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""The request related tools."""

import logging
import os
from collections import defaultdict
from typing import Any, DefaultDict, Optional

import urllib3
from requests import Session
from requests.adapters import HTTPAdapter
from requests.exceptions import RequestException, StreamConsumedError
from requests.models import PreparedRequest, Response
from tqdm import tqdm
from urllib3.util.retry import Retry

from tensorbay.client.log import RequestLogging, ResponseLogging
from tensorbay.exception import ResponseError

logger = logging.getLogger(__name__)


_CHUNK_SIZE = 8 * 1024


def _get_allowed_methods_keyword() -> str:
    splits = urllib3.__version__.split(".", 2)
    major = int(splits[0])
    minor = int(splits[1])
    return "allowed_methods" if (major, minor) >= (1, 26) else "method_whitelist"


# check the version of urllib3 and choose the correct keyword for "allowed_methods" in "Retry"
_ALLOWED_METHODS = _get_allowed_methods_keyword()


class Config:
    """This is a base class defining the concept of Request Config.

    Attributes:
        max_retries: Maximum retry times of the request.
        allowed_retry_methods: The allowed methods for retrying request.
        allowed_retry_status: The allowed status for retrying request.
            If both methods and status are fitted, the retrying strategy will work.
        timeout: Timeout value of the request in seconds.
        is_internal: Whether the request is from internal.

    """

    def __init__(self) -> None:

        self.max_retries = 3
        self.allowed_retry_methods = ["HEAD", "OPTIONS", "POST", "PUT"]
        self.allowed_retry_status = [429, 500, 502, 503, 504]

        self.timeout = 30
        self.is_internal = False
        self._x_source = "PYTHON-SDK"


config = Config()


class TimeoutHTTPAdapter(HTTPAdapter):
    """This class defines the http adapter for setting the timeout value.

    Arguments:
        *args: Extra arguments to initialize TimeoutHTTPAdapter.
        timeout: Timeout value of the post request in seconds.
        **kwargs: Extra keyword arguments to initialize TimeoutHTTPAdapter.

    """

    def __init__(self, *args: Any, timeout: Optional[int] = None, **kwargs: Any) -> None:
        self.timeout = timeout if timeout is not None else config.timeout
        super().__init__(*args, **kwargs)

    def send(  # pylint: disable=too-many-arguments
        self,
        request: PreparedRequest,
        stream: Any = False,
        timeout: Any = None,
        verify: Any = True,
        cert: Any = None,
        proxies: Any = None,
    ) -> Any:
        """Send the request.

        Arguments:
            request: The PreparedRequest being sent.
            stream: Whether to stream the request content.
            timeout: Timeout value of the post request in seconds.
            verify: A path string to a CA bundle to use or
                a boolean which controls whether to verify the server's TLS certificate.
            cert: User-provided SSL certificate.
            proxies: Proxies dict applying to the request.

        Returns:
            Response object.

        """
        if timeout is None:
            timeout = self.timeout
        return super().send(request, stream, timeout, verify, cert, proxies)


class UserSession(Session):
    """This class defines UserSession."""

    def __init__(self) -> None:
        super().__init__()
        # self.session.hooks["response"] = [logging_hook]

        retry_strategy = Retry(
            total=config.max_retries,
            status_forcelist=config.allowed_retry_status,
            raise_on_status=False,
            **{_ALLOWED_METHODS: config.allowed_retry_methods},
        )

        self.mount("http://", TimeoutHTTPAdapter(20, 20, retry_strategy))
        self.mount("https://", TimeoutHTTPAdapter(20, 20, retry_strategy))

    def request(  # type: ignore[override]
        self, method: str, url: str, *args: Any, **kwargs: Any
    ) -> Response:  # noqa: DAR401
        """Make the request.

        Arguments:
            method: Method for the request.
            url: URL for the request.
            *args: Extra arguments to make the request.
            **kwargs: Extra keyword arguments to make the request.

        Returns:
            Response of the request.

        Raises:
            ResponseError: If post response error.

        """
        try:
            response = super().request(method, url, *args, **kwargs)
            if response.status_code not in (200, 201):
                logger.error(
                    "Unexpected status code(%d)!%s", response.status_code, ResponseLogging(response)
                )
                raise ResponseError(response=response)

            logger.debug(ResponseLogging(response))
            return response

        except RequestException as error:
            logger.error(
                "%s.%s: %s%s",
                error.__class__.__module__,
                error.__class__.__name__,
                error,
                RequestLogging(error.request),
            )
            raise


SESSIONS: DefaultDict[int, UserSession] = defaultdict(UserSession)


def get_session() -> UserSession:
    """Create and return a session per PID so each sub-processes will use their own session.

    Returns:
        The session corresponding to the process.
    """
    return SESSIONS[os.getpid()]


class UserResponse:
    """This class used to read data from Response with stream method.

    Arguments:
        response: Response of the Session.request().

    """

    def __init__(self, response: Response):
        self.response = response

    def __enter__(self) -> "UserResponse":
        return self

    def __exit__(self, *_: Any) -> None:
        self.close()

    def close(self) -> None:
        """Close the response."""
        self.response.close()

    def read(self, amt: Optional[int] = None) -> bytes:
        """Read data from response.

        Arguments:
            amt: The needed read amount.

        Returns:
            Response of the request.

        """
        if amt is None:
            try:
                return b"".join(chunk for chunk in self.response.iter_content(_CHUNK_SIZE))

            except StreamConsumedError:
                return b""

        try:
            return next(self.response.iter_content(amt))  # type: ignore[no-any-return]

        except StopIteration:
            return b""


class Tqdm(tqdm):  # type: ignore[misc]
    """A wrapper class of tqdm for showing the process bar.

    Arguments:
        total: The number of excepted iterations.
        disable: Whether to disable the entire progress bar.

    """

    def __init__(self, total: int, disable: bool = False) -> None:
        super().__init__(desc="Uploading", total=total, disable=disable)

    def update_callback(self, _: Any) -> None:
        """Callback function for updating process bar when multithread task is done."""
        self.update()

    def update_for_skip(self, condition: bool) -> bool:
        """Update process bar for the items which are skipped in builtin filter function.

        Arguments:
            condition: The filter condition, the process bar will be updated if condition is False.

        Returns:
            The input condition.

        """
        if not condition:
            self.update()

        return condition



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of LeedsSportsPose dataset."""

import os

from tensorbay.dataset import Data, Dataset
from tensorbay.exception import ModuleImportError
from tensorbay.geometry import Keypoint2D
from tensorbay.label import LabeledKeypoints2D
from tensorbay.opendataset._utility import glob

DATASET_NAME = "LeedsSportsPose"


def LeedsSportsPose(path: str) -> Dataset:
    """`Leeds Sports Pose <http://sam.johnson.io/research/lsp.html>`_ dataset.

    The folder structure should be like::

        <path>
            joints.mat
            images/
                im0001.jpg
                im0002.jpg
                ...

    Arguments:
        path: The root directory of the dataset.

    Raises:
        ModuleImportError: When the module "scipy" can not be found.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    try:
        from scipy.io import loadmat  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as error:
        raise ModuleImportError(module_name=error.name) from error

    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    segment = dataset.create_segment()

    mat = loadmat(os.path.join(root_path, "joints.mat"))

    joints = mat["joints"].T
    image_paths = glob(os.path.join(root_path, "images", "*.jpg"))
    for image_path in image_paths:
        data = Data(image_path)
        data.label.keypoints2d = []
        index = int(os.path.basename(image_path)[2:6]) - 1  # get image index from "im0001.jpg"

        keypoints = LabeledKeypoints2D()
        for keypoint in joints[index]:
            keypoints.append(Keypoint2D(keypoint[0], keypoint[1], int(not keypoint[2])))

        data.label.keypoints2d.append(keypoints)
        segment.append(data)
    return dataset



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the DownsampledImagenet dataset."""

from tensorbay.opendataset.DownsampledImagenet.loader import DownsampledImagenet

__all__ = ["DownsampledImagenet"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""mixin classes for local file and remote file."""

import os
from hashlib import sha1
from typing import Any, Callable, Dict, Optional, Union
from urllib.parse import urljoin
from urllib.request import pathname2url

from _io import BufferedReader

from tensorbay.exception import ResponseError
from tensorbay.utility.repr import ReprMixin
from tensorbay.utility.requests import UserResponse, config, get_session


class URL:
    """URL is a class used to get and update the url.

    Arguments:
        url: The url.
        updater: A function used to update the url.

    """

    def __init__(self, url: str, updater: Callable[[], Optional[str]]) -> None:
        self._updater = updater
        self._getter: Callable[..., str] = lambda: url

    @classmethod
    def from_getter(cls, getter: Callable[..., str], updater: Callable[[], Optional[str]]) -> "URL":
        """Create a URL instance from the given getter and updater.

        Arguments:
            getter: The url getter of the file.
            updater: The updater of the url.

        Returns:
            The URL instance which stores the url and the updater.

        """
        obj: "URL" = object.__new__(cls)
        obj._getter = getter
        obj._updater = updater
        return obj

    def update(self) -> None:
        """Update the url."""
        url = self._updater()
        if url is not None:
            self._getter = lambda: url  # type: ignore[assignment, return-value]

    def get(self) -> str:
        """Get the url of the file.

        Returns:
            The url.

        """
        return self._getter()


class FileMixin(ReprMixin):
    """FileMixin is a mixin class to mixin file related methods for local file.

    Arguments:
        local_path: The file local path.

    Attributes:
        path: The file local path.

    """

    _checksum: str

    _repr_maxlevel = 3
    _BUFFER_SIZE = 65536

    def __init__(self, local_path: str) -> None:
        self.path = local_path

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.path}")'

    def _get_callback_body(self) -> Dict[str, Any]:
        return {"checksum": self.get_checksum(), "fileSize": os.path.getsize(self.path)}

    def get_checksum(self) -> str:
        """Get and cache the sha1 checksum of the local data.

        Returns:
            The sha1 checksum of the local data.

        """
        if not hasattr(self, "_checksum"):
            sha1_object = sha1()
            with open(self.path, "rb") as fp:
                while True:
                    data = fp.read(self._BUFFER_SIZE)
                    if not data:
                        break
                    sha1_object.update(data)

            self._checksum = sha1_object.hexdigest()

        return self._checksum

    def get_url(self) -> str:
        """Return the url of the local data file.

        Returns:
            The url of the local data.

        """
        return urljoin("file:", pathname2url(os.path.abspath(self.path)))

    def open(self) -> BufferedReader:
        """Return the binary file pointer of this file.

        The local file pointer will be obtained by build-in ``open()``.

        Returns:
            The local file pointer for this data.

        """
        return open(self.path, "rb")


class RemoteFileMixin(ReprMixin):
    """RemoteFileMixin is a mixin class to mixin file related methods for remote file.

    Arguments:
        local_path: The file local path.
        url: The URL instance used to get and update url.
        cache_path: The path to store the cache.

    Attributes:
        path: The file local path.

    """

    _repr_maxlevel = 3

    def __init__(
        self,
        remote_path: str,
        *,
        url: Optional[URL] = None,
        cache_path: str = "",
    ) -> None:
        self.path = remote_path
        self.url = url
        self.cache_path = os.path.join(cache_path, remote_path) if cache_path else ""

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.path}")'

    def _urlopen(self) -> UserResponse:

        url = self.get_url()
        if not self.url:
            raise ValueError(f"The file cannot open because {self._repr_head()} has no url")

        try:
            session = get_session()
            return UserResponse(session.request("GET", url, timeout=config.timeout, stream=True))
        except ResponseError as error:
            if error.response.status_code == 403:
                self.url.update()
                return UserResponse(
                    get_session().request("GET", url, timeout=config.timeout, stream=True)
                )
            raise

    def get_url(self) -> str:
        """Return the url of the data hosted by tensorbay.

        Returns:
            The url of the data.

        Raises:
            ValueError: When the url is missing.

        """
        if not self.url:
            raise ValueError(f"The file URL cannot be got because {self._repr_head()} has no url")

        return self.url.get()

    def open(self) -> Union[UserResponse, BufferedReader]:
        """Return the binary file pointer of this file.

        The remote file pointer will be obtained by ``urllib.request.urlopen()``.

        Returns:
            The remote file pointer for this data.

        """
        cache_path = self.cache_path
        if not cache_path:
            return self._urlopen()

        if not os.path.exists(cache_path):
            dirname = os.path.dirname(cache_path)
            os.makedirs(dirname, exist_ok=True)

            with self._urlopen() as fp:
                with open(cache_path, "wb") as cache:
                    cache.write(fp.read())

        return open(cache_path, "rb")



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the Halpe Full-Body Human Keypoints and HOI-Det dataset."""

from tensorbay.opendataset.HalpeFullBody.loader import HalpeFullBody

__all__ = ["HalpeFullBody"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of AnimalsWithAttributes2 dataset."""

import os

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification
from tensorbay.opendataset._utility import glob

DATASET_NAME = "AnimalsWithAttributes2"


def AnimalsWithAttributes2(path: str) -> Dataset:
    """`Animals with attributes 2 <https://cvml.ist.ac.at/AwA2/>`_ dataset.

    The file structure should be like::

        <path>
            classes.txt
            predicates.txt
            predicate-matrix-binary.txt
            JPEGImages/
                <classname>/
                    <imagename>.jpg
                ...
            ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    segment = dataset.create_segment()

    with open(os.path.join(root_path, "classes.txt"), encoding="utf-8") as fp:
        class_names = [line[:-1].split("\t", 1)[-1] for line in fp]

    with open(os.path.join(root_path, "predicates.txt"), encoding="utf-8") as fp:
        attribute_keys = [line[:-1].split("\t", 1)[-1] for line in fp]

    with open(os.path.join(root_path, "predicate-matrix-binary.txt"), encoding="utf-8") as fp:
        attribute_values = [line[:-1].split(" ") for line in fp]

    attribute_mapping = {}
    for class_name, values in zip(class_names, attribute_values):
        attribute_mapping[class_name] = Classification(
            category=class_name,
            attributes=dict(zip(attribute_keys, (bool(int(value)) for value in values))),
        )

    for class_name in sorted(os.listdir(os.path.join(root_path, "JPEGImages"))):
        image_paths = glob(os.path.join(root_path, "JPEGImages", class_name, "*.jpg"))
        label = attribute_mapping[class_name]
        for image_path in image_paths:
            data = Data(image_path)
            data.label.classification = label
            segment.append(data)

    return dataset



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Dataset related classes."""

from tensorbay.dataset.data import AuthData, Data, RemoteData
from tensorbay.dataset.dataset import Dataset, FusionDataset, Notes
from tensorbay.dataset.frame import Frame
from tensorbay.dataset.segment import FusionSegment, Segment

__all__ = [
    "AuthData",
    "Data",
    "Dataset",
    "Frame",
    "FusionDataset",
    "FusionSegment",
    "Notes",
    "RemoteData",
    "Segment",
]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""LabeledKeypoints2D, Keypoints2DSubcatalog.

:class:`Keypoints2DSubcatalog` defines the subcatalog for 2D keypoints type of labels.

:class:`LabeledKeypoints2D` is the 2D keypoints type of label,
which is often used for CV tasks such as human body pose estimation.

"""

from typing import Any, Dict, Iterable, List, Optional, Type, TypeVar, Union

from tensorbay.geometry import Keypoints2D
from tensorbay.label.basic import SubcatalogBase, _LabelBase
from tensorbay.label.supports import (
    AttributesMixin,
    CategoriesMixin,
    IsTrackingMixin,
    KeypointsInfo,
)
from tensorbay.utility import ReprType, attr, attr_base, common_loads


class Keypoints2DSubcatalog(SubcatalogBase, IsTrackingMixin, CategoriesMixin, AttributesMixin):
    """This class defines the subcatalog for 2D keypoints type of labels.

    Arguments:
        is_tracking: A boolean value indicates whether the corresponding
            subcatalog contains tracking information.

    Attributes:
        description: The description of the entire 2D keypoints subcatalog.
        categories: All the possible categories in the corresponding dataset
            stored in a :class:`~tensorbay.utility.name.NameList`
            with the category names as keys
            and the :class:`~tensorbay.label.supports.CategoryInfo` as values.
        category_delimiter: The delimiter in category values indicating parent-child relationship.
        attributes: All the possible attributes in the corresponding dataset
            stored in a :class:`~tensorbay.utility.name.NameList`
            with the attribute names as keys
            and the :class:`~tensorbay.label.attribute.AttributeInfo` as values.
        is_tracking: Whether the Subcatalog contains tracking information.

    Examples:
        *Initialization Method 1:* Init from ``Keypoints2DSubcatalog.loads()`` method.

        >>> catalog = {
        ...     "KEYPOINTS2D": {
        ...         "isTracking": True,
        ...         "categories": [{"name": "0"}, {"name": "1"}],
        ...         "attributes": [{"name": "gender", "enum": ["male", "female"]}],
        ...         "keypoints": [
        ...             {
        ...                 "number": 2,
        ...                  "names": ["L_shoulder", "R_Shoulder"],
        ...                  "skeleton": [(0, 1)],
        ...             }
        ...         ],
        ...     }
        ... }
        >>> Keypoints2DSubcatalog.loads(catalog["KEYPOINTS2D"])
        Keypoints2DSubcatalog(
          (is_tracking): True,
          (keypoints): [...],
          (categories): NameList [...],
          (attributes): NameList [...]
        )

        *Initialization Method 2:* Init an empty Keypoints2DSubcatalog and then add the attributes.

        >>> from tensorbay.label import CategoryInfo, AttributeInfo, KeypointsInfo
        >>> from tensorbay.utility import NameList
        >>> categories = NameList()
        >>> categories.append(CategoryInfo("a"))
        >>> attributes = NameList()
        >>> attributes.append(AttributeInfo("gender", enum=["female", "male"]))
        >>> keypoints2d_subcatalog = Keypoints2DSubcatalog()
        >>> keypoints2d_subcatalog.is_tracking = True
        >>> keypoints2d_subcatalog.categories = categories
        >>> keypoints2d_subcatalog.attributes = attributes
        >>> keypoints2d_subcatalog.add_keypoints(
        ...     2,
        ...     names=["L_shoulder", "R_Shoulder"],
        ...     skeleton=[(0,1)],
        ...     visible="BINARY",
        ...     parent_categories="shoulder",
        ...     description="12345",
        ... )
        >>> keypoints2d_subcatalog
        Keypoints2DSubcatalog(
          (is_tracking): True,
          (keypoints): [...],
          (categories): NameList [...],
          (attributes): NameList [...]
        )

    """

    _keypoints: List[KeypointsInfo] = attr(key="keypoints")

    def __init__(self, is_tracking: bool = False) -> None:
        SubcatalogBase.__init__(self)
        IsTrackingMixin.__init__(self, is_tracking)
        self._keypoints: List[KeypointsInfo] = []

    @property
    def keypoints(self) -> List[KeypointsInfo]:
        """Return the KeypointsInfo of the Subcatalog.

        Returns:
            A list of :class:`~tensorbay.label.supports.KeypointsInfo`.

        Examples:
            >>> keypoints2d_subcatalog = Keypoints2DSubcatalog()
            >>> keypoints2d_subcatalog.add_keypoints(2)
            >>> keypoints2d_subcatalog.keypoints
            [KeypointsInfo(
              (number): 2
            )]

        """
        return self._keypoints

    def add_keypoints(
        self,
        number: int,
        *,
        names: Optional[Iterable[str]] = None,
        skeleton: Optional[Iterable[Iterable[int]]] = None,
        visible: Optional[str] = None,
        parent_categories: Union[None, str, Iterable[str]] = None,
        description: str = "",
    ) -> None:
        """Add a type of keypoints to the subcatalog.

        Arguments:
            number: The number of keypoints.
            names: All the names of keypoints.
            skeleton: The skeleton of the keypoints
                indicating which keypoint should connect with another.
            visible: The visible type of the keypoints, can only be 'BINARY' or 'TERNARY'.
                It determines the range of the
                :attr:`Keypoint2D.v<tensorbay.geometry.keypoint.Keypoint2D.v>`.
            parent_categories: The parent categories of the keypoints.
            description: The description of keypoints.

        Examples:
            >>> keypoints2d_subcatalog = Keypoints2DSubcatalog()
            >>> keypoints2d_subcatalog.add_keypoints(
            ...     2,
            ...     names=["L_shoulder", "R_Shoulder"],
            ...     skeleton=[(0,1)],
            ...     visible="BINARY",
            ...     parent_categories="shoulder",
            ...     description="12345",
            ... )
            >>> keypoints2d_subcatalog.keypoints
            [KeypointsInfo(
              (number): 2,
              (names): [...],
              (skeleton): [...],
              (visible): 'BINARY',
              (parent_categories): [...]
            )]

        """
        self._keypoints.append(
            KeypointsInfo(
                number=number,
                names=names,
                skeleton=skeleton,
                visible=visible,
                parent_categories=parent_categories,
                description=description,
            )
        )

    def dumps(self) -> Dict[str, Any]:
        """Dumps all the information of the keypoints into a dict.

        Returns:
            A dict containing all the information of this Keypoints2DSubcatalog.

        Examples:
            >>> # keypoints2d_subcatalog is the instance initialized above.
            >>> keypoints2d_subcatalog.dumps()
            {
                'isTracking': True,
                'categories': [{'name': 'a'}],
                'attributes': [{'name': 'gender', 'enum': ['female', 'male']}],
                'keypoints': [
                    {
                        'number': 2,
                        'names': ['L_shoulder', 'R_Shoulder'],
                        'skeleton': [(0, 1)],
                    }
                ]
            }

        """
        return self._dumps()


class LabeledKeypoints2D(_LabelBase, Keypoints2D):
    """This class defines the concept of 2D keypoints label.

    :class:`LabeledKeypoints2D` is the 2D keypoints type of label,
    which is often used for CV tasks such as human body pose estimation.

    Arguments:
        keypoints: A list of 2D keypoint.
        category: The category of the label.
        attributes: The attributes of the label.
        instance: The instance id of the label.

    Attributes:
        category: The category of the label.
        attributes: The attributes of the label.
        instance: The instance id of the label.

    Examples:
        >>> LabeledKeypoints2D(
        ...     [(1, 2), (2, 3)],
        ...     category="example",
        ...     attributes={"key": "value"},
        ...     instance="123",
        ... )
        LabeledKeypoints2D [
          Keypoint2D(1, 2),
          Keypoint2D(2, 3)
        ](
          (category): 'example',
          (attributes): {...},
          (instance): '123'
        )

    """

    _T = TypeVar("_T", bound="LabeledKeypoints2D")

    _repr_type = ReprType.SEQUENCE
    _repr_attrs = _LabelBase._repr_attrs
    _attrs_base: Keypoints2D = attr_base(key="keypoints2d")

    def __init__(
        self,
        keypoints: Optional[Iterable[Iterable[float]]] = None,
        *,
        category: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        instance: Optional[str] = None,
    ) -> None:
        Keypoints2D.__init__(self, keypoints)  # type: ignore[arg-type]
        _LabelBase.__init__(self, category, attributes, instance)

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:  # type: ignore[override]
        """Loads a LabeledKeypoints2D from a dict containing the information of the label.

        Arguments:
            contents: A dict containing the information of the 2D keypoints label.

        Returns:
            The loaded :class:`LabeledKeypoints2D` object.

        Examples:
            >>> contents = {
            ...     "keypoints2d": [
            ...         {"x": 1, "y": 1, "v": 2},
            ...         {"x": 2, "y": 2, "v": 2},
            ...     ],
            ...     "category": "example",
            ...     "attributes": {"key": "value"},
            ...     "instance": "12345",
            ... }
            >>> LabeledKeypoints2D.loads(contents)
            LabeledKeypoints2D [
              Keypoint2D(1, 1, 2),
              Keypoint2D(2, 2, 2)
            ](
              (category): 'example',
              (attributes): {...},
              (instance): '12345'
            )

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:  # type: ignore[override]
        """Dumps the current 2D keypoints label into a dict.

        Returns:
            A dict containing all the information of the 2D keypoints label.

        Examples:
            >>> labeledkeypoints2d = LabeledKeypoints2D(
            ...     [(1, 1, 2), (2, 2, 2)],
            ...     category="example",
            ...     attributes={"key": "value"},
            ...     instance="123",
            ... )
            >>> labeledkeypoints2d.dumps()
            {
                'category': 'example',
                'attributes': {'key': 'value'},
                'instance': '123',
                'keypoints2d': [{'x': 1, 'y': 1, 'v': 2}, {'x': 2, 'y': 2, 'v': 2}],
            }

        """
        return self._dumps()



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas ls."""

from typing import Iterable, Optional

import click

from tensorbay.cli.tbrn import TBRN, TBRNType
from tensorbay.cli.utility import ContextInfo, error, exception_handler, get_dataset_client
from tensorbay.client import GAS
from tensorbay.client.dataset import FusionDatasetClient


def _echo_data(
    dataset_name: str,
    draft_number: Optional[int],
    revision: Optional[str],
    segment_name: str,
    data_iter: Iterable[str],
) -> None:
    """Echo files in data_iter under 'tb:dataset_name:segment_name'.

    Arguments:
        dataset_name: The name of the dataset the segment belongs to.
        draft_number: The draft number (if the status is draft).
        revision: The revision (if the status is not draft).
        segment_name: The name of the segment.
        data_iter: Iterable data to be echoed.

    """
    for data in data_iter:
        click.echo(
            TBRN(
                dataset_name,
                segment_name,
                remote_path=data,
                draft_number=draft_number,
                revision=revision,
            ).get_tbrn()
        )


def _ls_dataset(gas: GAS, tbrn_info: TBRN, list_all_files: bool, show_total_num: bool) -> None:
    dataset_client = get_dataset_client(gas, tbrn_info)

    segment_names = dataset_client.list_segment_names()
    if not list_all_files:
        if show_total_num:
            click.echo(f"total {len(segment_names)}")
        for segment_name in segment_names:
            click.echo(TBRN(tbrn_info.dataset_name, segment_name).get_tbrn())
        return

    if isinstance(dataset_client, FusionDatasetClient):
        error('"-a" flag is not supported for fusion dataset yet')

    segment_paths = [
        dataset_client.get_segment(segment_name).list_data_paths() for segment_name in segment_names
    ]
    if show_total_num:
        total = sum(len(segment_path) for segment_path in segment_paths)
        click.echo(f"total {total}")

    for segment_name, segment_path in zip(segment_names, segment_paths):
        _echo_data(
            tbrn_info.dataset_name,
            tbrn_info.draft_number,
            tbrn_info.revision,
            segment_name,
            segment_path,
        )


def _ls_segment(
    gas: GAS,
    tbrn_info: TBRN,
    list_all_files: bool,  # pylint: disable=unused-argument
    show_total_num: bool,
) -> None:
    dataset_client = get_dataset_client(gas, tbrn_info)
    if isinstance(dataset_client, FusionDatasetClient):
        error("List fusion segment is not supported yet")

    segment_path = dataset_client.get_segment(tbrn_info.segment_name).list_data_paths()
    if show_total_num:
        click.echo(f"total {len(segment_path)}")
    _echo_data(
        tbrn_info.dataset_name,
        tbrn_info.draft_number,
        tbrn_info.revision,
        tbrn_info.segment_name,
        segment_path,
    )


def _ls_normal_file(
    gas: GAS,
    tbrn_info: TBRN,
    list_all_files: bool,  # pylint: disable=unused-argument
    show_total_num: bool,  # pylint: disable=unused-argument
) -> None:
    dataset_client = get_dataset_client(gas, tbrn_info)
    if isinstance(dataset_client, FusionDatasetClient):
        error("List data in fusion segment is not supported yet")

    remote_path = tbrn_info.remote_path
    dataset_client.get_segment(tbrn_info.segment_name).get_data(remote_path=remote_path)
    _echo_data(
        tbrn_info.dataset_name,
        tbrn_info.draft_number,
        tbrn_info.revision,
        tbrn_info.segment_name,
        (remote_path,),
    )


_LS_FUNCS = {
    TBRNType.DATASET: _ls_dataset,
    TBRNType.SEGMENT: _ls_segment,
    TBRNType.NORMAL_FILE: _ls_normal_file,
}


@exception_handler
def _implement_ls(obj: ContextInfo, tbrn: str, list_all_files: bool, show_total_num: bool) -> None:
    gas = obj.get_gas()
    if not tbrn:
        dataset_names = gas.list_dataset_names()
        if show_total_num:
            click.echo(f"total {len(dataset_names)}")
        for dataset_name in dataset_names:
            click.echo(TBRN(dataset_name).get_tbrn())
        return

    tbrn_info = TBRN(tbrn=tbrn)
    _LS_FUNCS[tbrn_info.type](gas, tbrn_info, list_all_files, show_total_num)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas log."""

import bisect
import sys
from collections import defaultdict
from datetime import datetime
from functools import partial
from itertools import cycle, islice, zip_longest
from textwrap import indent
from typing import (
    Callable,
    DefaultDict,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

import click

from tensorbay.cli.auth import INDENT
from tensorbay.cli.tbrn import TBRN, TBRNType
from tensorbay.cli.utility import ContextInfo, error, exception_handler, shorten
from tensorbay.client.gas import DatasetClientType
from tensorbay.client.struct import ROOT_COMMIT_ID, Commit, Draft

_LEFT_BRACKET = click.style("(", fg="yellow", reset=False)
_COMMA = click.style(", ", fg="yellow", reset=False)
_RIGHT_BRACKET = click.style(")", fg="yellow", reset=True)

_FULL_COMMIT_MESSAGE = f"""{click.style("commit {}", fg="yellow")}
Author: {{}}
Date: {{}}

    {{}}
"""

_FULL_DRAFT_MESSAGE = f"""{click.style("draft {}", fg="red")}
Date: {{}}

    {{}}
"""

_ONELINE_COMMIT_MESSAGE = f"""{click.style("{}", fg="yellow")} {{}}
"""

_ONELINE_DRAFT_MESSAGE = f"""{click.style("{}", fg="red")} {{}}
"""

_LOG_COLORS = (
    "red",
    "green",
    "blue",
    "magenta",
    "cyan",
    "bright_blue",
    "bright_magenta",
    "bright_cyan",
    "bright_red",
    "bright_green",
    "bright_yellow",
)
_R = TypeVar("_R", bound="_RootCommitNode")
_D = TypeVar("_D", bound="_DraftNode")
_T = Union["_RootCommitNode", "_DraftNode"]


@exception_handler
def _implement_log(  # pylint: disable=too-many-arguments
    obj: ContextInfo,
    tbrn: str,
    max_count: Optional[int],
    oneline: bool,
    is_all: bool,
    graph: bool,
    show_drafts: bool,
) -> None:
    gas = obj.get_gas()
    tbrn_info = TBRN(tbrn=tbrn)
    if tbrn_info.type != TBRNType.DATASET:
        error(f'To log commits, "{tbrn}" must be a dataset')

    dataset_client = gas._get_dataset_with_any_type(  # pylint: disable=protected-access
        tbrn_info.dataset_name
    )
    commit_id_to_branches: DefaultDict[str, List[str]] = defaultdict(list)
    for branch in dataset_client.list_branches():
        commit_id_to_branches[branch.commit_id].append(branch.name)
    if is_all:
        revisions: List[Optional[str]] = [branch.name for branch in dataset_client.list_branches()]
    else:
        revisions = (
            [tbrn_info.revision] if tbrn_info.revision else [dataset_client.status.branch_name]
        )
    Printer: Union[Type[_GraphPrinter], Type[_Printer]] = _GraphPrinter if graph else _Printer
    message_generator = islice(
        Printer(
            dataset_client, revisions, commit_id_to_branches, oneline, show_drafts=show_drafts
        ).generate_commits_and_drafts_messages(),
        max_count,
    )
    _echo_messages(message_generator)


def _join_branch_names(commit_id: str, branch_names: List[str]) -> str:
    return (
        f"{commit_id} {_LEFT_BRACKET}"
        f"{_COMMA.join(click.style(name, fg='green', reset=False) for name in branch_names)}"
        f"{_RIGHT_BRACKET}"
    )


def _get_oneline_commit_message(commit: Commit, branch_names: Optional[List[str]]) -> str:
    commit_id = shorten(commit.commit_id)
    if branch_names:
        commit_id = _join_branch_names(commit_id, branch_names)
    return _ONELINE_COMMIT_MESSAGE.format(commit_id, commit.title)


def _get_oneline_draft_message(draft: Draft) -> str:
    return _ONELINE_DRAFT_MESSAGE.format(draft.number, draft.title)


def _get_full_commit_message(commit: Commit, branch_names: Optional[List[str]]) -> str:
    description = commit.description
    if description:
        description = f"\n\n{indent(description, INDENT)}"
    commit_message = f"{commit.title}{description}\n"
    commit_id = commit.commit_id
    if branch_names:
        commit_id = _join_branch_names(commit_id, branch_names)
    return _FULL_COMMIT_MESSAGE.format(
        commit_id,
        commit.committer.name,
        datetime.fromtimestamp(commit.committer.date).strftime("%a %b %d %H:%M:%S %y"),
        commit_message,
    )


def _get_full_draft_message(draft: Draft) -> str:
    description = draft.description
    if description:
        description = f"\n\n{indent(description, INDENT)}"
    draft_message = f"{draft.title}{description}\n"
    return _FULL_DRAFT_MESSAGE.format(
        draft.number,
        datetime.fromtimestamp(draft.updated_at).strftime("%a %b %d %H:%M:%S %y"),
        draft_message,
    )


class _Printer:
    """This class defines the structure of logging commits and open drafts.

    Arguments:
        dataset_client: The dataset that needs to be logged.
        revisions: The revisions that needs to be logged.
        oneline: Whether to log with oneline method.
        commit_id_to_branches: The map of commit id to branch name.
        show_drafts: Whether to log open drafts.

    """

    def __init__(
        self,
        dataset_client: DatasetClientType,
        revisions: List[Optional[str]],
        commit_id_to_branches: Dict[str, List[str]],
        oneline: bool,
        *,
        show_drafts: bool,
    ):
        all_commit_logs = list(map(dataset_client.list_commits, revisions))
        all_drafts: List[Draft] = []
        error_message = f'Dataset "{dataset_client.name}" has no commit history'
        if show_drafts:
            error_message += " or open drafts"
            for revision in revisions:
                all_drafts.extend(dataset_client.list_drafts(branch_name=revision))

        if not all_commit_logs[0]:
            if not all_drafts:
                error(error_message)
            self._sorted_commit_logs = []
        else:
            # Sort logs from different branches by the date of the latest commit of each branch.
            self._sorted_commit_logs = sorted(all_commit_logs, key=lambda x: x[0].committer.date)

        self._commit_id_to_branches = commit_id_to_branches
        self._commit_printer, self._draft_printer = (
            (_get_oneline_commit_message, _get_oneline_draft_message)
            if oneline
            else (_get_full_commit_message, _get_full_draft_message)
        )

        self._sorted_drafts = sorted(all_drafts, key=lambda x: x.updated_at)
        self._keys = [log[0].committer.date for log in self._sorted_commit_logs]

    def _merge(self, latest_commit: Commit) -> bool:
        if len(self._sorted_commit_logs) <= 1:
            return False

        date = latest_commit.committer.date
        commit_id = latest_commit.commit_id
        for log in islice(reversed(self._sorted_commit_logs), 1, None):
            commit = log[0]
            # Traverse all the commits with the same timestamp,
            # if the commit id is the same as the latest commit,
            # then merge the branch where the latest commit is located.
            if commit.committer.date != date:
                return False
            # Same commit, merge branches.
            if commit.commit_id == commit_id:
                break
        else:
            return False
        # Merge branches.
        del self._sorted_commit_logs[-1]
        del self._keys[-1]
        return True

    def _sort(self) -> None:
        """Sort commits paging list by commit date."""
        # Only one branch exists.
        if len(self._sorted_commit_logs) == 1:
            return

        # Binary insert.
        log = self._sorted_commit_logs.pop()
        del self._keys[-1]
        date = log[0].committer.date
        index = bisect.bisect_left(self._keys, date)
        self._sorted_commit_logs.insert(index, log)
        self._keys.insert(index, date)

    def generate_commits_and_drafts_messages(self) -> Iterator[str]:
        """Get the messages of all commits and open drafts.

        Yields:
            The commit or draft message.

        """
        latest_draft: Optional[Draft] = None
        while True:
            try:
                latest_commit = self._sorted_commit_logs[-1][0]
            except IndexError:
                if latest_draft:
                    yield self._draft_printer(latest_draft)
                if self._sorted_drafts:
                    yield from map(self._draft_printer, reversed(self._sorted_drafts))
                return
            if not latest_draft and self._sorted_drafts:
                latest_draft = self._sorted_drafts.pop()
            if latest_draft and latest_draft.updated_at > latest_commit.committer.date:
                yield self._draft_printer(latest_draft)
                latest_draft = None
            else:
                del self._sorted_commit_logs[-1][0]
                if self._merge(latest_commit):
                    continue
                yield self._commit_printer(
                    latest_commit, self._commit_id_to_branches.get(latest_commit.commit_id)
                )
                self._sort()


class _RootCommitNode:  # pylint: disable=too-many-instance-attributes
    SIGN = "*"

    def __init__(self) -> None:
        self.parent: Optional[_RootCommitNode] = None
        self.date = 0
        self.key = ROOT_COMMIT_ID
        self.available_child_num = 0
        self.get_oneline_message: Callable[
            [List[str]], str
        ] = lambda _: _ONELINE_COMMIT_MESSAGE.format(shorten(ROOT_COMMIT_ID), "ROOT_COMMIT")
        self.get_full_message: Callable[
            [List[str]], str
        ] = lambda _: f'{click.style("ROOT_COMMIT {}", fg="yellow")}'.format(self.key)

    def add_child(self, child_node: _T) -> None:
        """Save the child node for the parent node.

        Arguments:
            child_node: The child node.

        """
        child_node.parent = self
        self.available_child_num += 1


class _CommitNode(_RootCommitNode):
    """This class defines the tree struct of graphical logging commits.

    Arguments:
        commit: The commit that needs to be saved in the tree.

    """

    def __init__(self, commit: Commit):
        super().__init__()
        self.commit = commit
        self.date = commit.committer.date
        self.key = commit.commit_id
        self.get_oneline_message = partial(_get_oneline_commit_message, commit)
        self.get_full_message = partial(_get_full_commit_message, commit)


class _DraftNode:  # pylint: disable=too-many-instance-attributes
    """This class defines the tree struct of graphical logging drafts.

    Arguments:
        draft: The draft that needs to be saved in the tree.

    """

    SIGN = "#"

    def __init__(self, draft: Draft) -> None:
        self.draft = draft
        self.parent: Optional[_CommitNode] = None
        self.date = draft.updated_at
        self.available_child_num = 0
        self.key = str(draft.number)
        self.get_oneline_message: Callable[[List[str]], str] = lambda _: _get_oneline_draft_message(
            draft
        )
        self.get_full_message: Callable[[List[str]], str] = lambda _: _get_full_draft_message(draft)


class _GraphPrinter:
    """This class defines the structure of logging graphical commits and open drafts stack.

    Arguments:
        dataset_client: The dataset that needs to be logged.
        revisions: The revisions that needs to be logged.
        oneline: Whether to log with oneline method.
        commit_id_to_branches: The map of commit id to branch name.
        show_drafts: Whether to log open drafts.

    """

    def __init__(
        self,
        dataset_client: DatasetClientType,
        revisions: List[Optional[str]],
        commit_id_to_branches: Dict[str, List[str]],
        oneline: bool,
        *,
        show_drafts: bool,
    ):
        self._key_to_branches = commit_id_to_branches
        self._graph_printer = self._add_oneline_graph if oneline else self._add_full_graph

        self._sorted_leaves = self._build_tree(dataset_client, revisions, show_drafts)
        error_message = f'Dataset "{dataset_client.name}" has no commit history'
        if show_drafts:
            error_message += " or open drafts"
        if not self._sorted_leaves:
            error(error_message)
        self._pointer = 0
        self._merge_pointer: Optional[int] = None
        self._log_colors = cycle(_LOG_COLORS)
        self._layer_colors: List[str] = [next(self._log_colors)]

    def _get_log_node(self) -> _T:
        """Get the next log node.

        Returns:
            The next log node.

        Raises:
            RuntimeError: Graphical logging algorithm error.

        """
        self._merge_pointer = None
        for log_node in islice(self._sorted_leaves, self._pointer, None):
            if log_node.available_child_num == 0:
                return log_node

            self._pointer += 1
            if self._pointer >= len(self._layer_colors):
                self._layer_colors.append(next(self._log_colors))

        raise RuntimeError("Graphical logging algorithm error.")

    def _merge_branches(self, parent: _RootCommitNode) -> None:
        """Merge branches.

        Arguments:
            parent: The parent node.

        """
        index = self._sorted_leaves.index(parent)
        self._pointer, self._merge_pointer = sorted((index, self._pointer))
        del self._sorted_leaves[self._merge_pointer]

    def _set_next_node(self, node: _RootCommitNode) -> None:
        """Set the next node at the position of the printing pointer.

        Arguments:
            node: The next node of the printing pointer.

        """
        node.available_child_num -= 1
        self._sorted_leaves[self._pointer] = node

    def _add_oneline_graph(self, node: _T, original_pointer: int) -> str:
        message = node.get_oneline_message(self._key_to_branches[node.key])
        prefixes = self._get_colorful_prefixes()
        # Don't merge branches.
        if self._merge_pointer is None:
            return f"{self._get_title_prefix(prefixes, original_pointer, node.SIGN)}{message}"

        # Merge branches.
        del self._layer_colors[self._merge_pointer]
        lines = [f"{self._get_title_prefix(prefixes, original_pointer, node.SIGN)}{message}"]
        lines.extend(
            f"{prefixes}\n"
            for prefixes in self._get_merge_prefixes(
                self._get_colorful_prefixes(), self._merge_pointer, self._pointer
            )
        )
        return "".join(lines)

    def _add_full_graph(self, node: _T, original_pointer: int) -> str:
        message = node.get_full_message(self._key_to_branches[node.key])
        splitlines = iter(message.splitlines())
        original_prefixes = self._get_colorful_prefixes()
        lines = [
            f"{self._get_title_prefix(original_prefixes, original_pointer, node.SIGN)}"
            f"{next(splitlines)}\n"
        ]
        # Don't merge branches.
        if self._merge_pointer is None:
            details_prefix = "".join(original_prefixes)
            lines.extend(f"{details_prefix}{line}\n" for line in splitlines)
        else:
            # Merge branches.
            del self._layer_colors[self._merge_pointer]
            prefixes = self._get_colorful_prefixes()
            merge_prefixes = self._get_merge_prefixes(prefixes, self._merge_pointer, self._pointer)
            lines.extend(self._combine_details(list(merge_prefixes), list(splitlines), prefixes))
        return "".join(lines)

    def _get_merge_prefixes(
        self, prefixes: List[str], merge_pointer: int, pointer: int
    ) -> Iterator[str]:
        for i in range(merge_pointer, pointer, -1):
            temp_prefixes = prefixes.copy()
            temp_prefixes[2 * i - 1] = click.style("/", fg=self._layer_colors[pointer])
            if i == merge_pointer:
                for j, color in enumerate(self._layer_colors[i:], i):
                    temp_prefixes[2 * j] = " "
                    temp_prefixes[2 * j + 1] = click.style("/", fg=color)
            yield "".join(temp_prefixes)

    def _get_colorful_prefixes(self) -> List[str]:
        prefixes = []
        for color in self._layer_colors:
            prefixes.append(click.style("|", fg=color))
            prefixes.append(" ")
        return prefixes

    @staticmethod
    def _get_title_prefix(prefixes: List[str], original_pointer: int, sign: str) -> str:
        title_prefixes = prefixes.copy()
        title_prefixes[2 * original_pointer] = sign
        return "".join(title_prefixes)

    @staticmethod
    def _combine_details(
        merge_prefixes: List[str], messages: List[str], fill_prefix: List[str]
    ) -> Iterator[str]:
        fillvalue = "" if len(merge_prefixes) > len(messages) else "".join(fill_prefix)
        for prefix, message in zip_longest(merge_prefixes, messages, fillvalue=fillvalue):
            yield f"{prefix}  {message}\n"

    def _build_tree(
        self,
        dataset_client: DatasetClientType,
        revisions: List[Optional[str]],
        show_drafts: bool,
    ) -> List[_T]:
        commit_to_node: Dict[str, _RootCommitNode] = {}
        leaves: Dict[str, _T] = {}
        for revision in revisions:
            child_node: Optional[_RootCommitNode] = None
            for commit in dataset_client.list_commits(revision):
                commit_id = commit.commit_id
                current_node = commit_to_node.get(commit_id, _CommitNode(commit))
                if child_node:
                    current_node.add_child(child_node)
                # Commit already exists in the tree.
                if commit_id in commit_to_node:
                    break

                # Save leaf node to leaf set.
                if not child_node:
                    leaves[commit_id] = current_node
                # Save commit to commit dict.
                commit_to_node[commit_id] = current_node
                child_node = current_node
            if show_drafts:
                for draft in dataset_client.list_drafts(branch_name=revision):
                    draft_node = _DraftNode(draft)
                    leaves[draft_node.key] = draft_node
                    self._key_to_branches[draft_node.key] = [draft.branch_name]
                    parent_commit_id = draft.parent_commit_id
                    if parent_commit_id in commit_to_node:
                        parent_node = commit_to_node[parent_commit_id]
                    else:
                        parent_node = _RootCommitNode()
                        commit_to_node[parent_commit_id] = parent_node
                        if child_node:
                            parent_node.add_child(child_node)
                    parent_node.add_child(draft_node)

        return self._check_and_sort_leaves(leaves)

    @staticmethod
    def _check_and_sort_leaves(leaves: Dict[str, _T]) -> List[_T]:
        # Check the correction of leaf set.
        delete_keys = [
            key for key, leaf_node in leaves.items() if leaf_node.available_child_num != 0
        ]
        for key in delete_keys:
            del leaves[key]
        return sorted(leaves.values(), key=lambda x: x.date, reverse=True)

    def generate_commits_and_drafts_messages(self) -> Iterator[str]:
        """Get the graphical message.

        Yields:
            The graphical message.

        """
        change_color = False
        while True:
            current_node = self._get_log_node()
            parent = current_node.parent
            original_pointer = self._pointer
            if parent not in self._sorted_leaves:
                if change_color:
                    # pylint: disable=stop-iteration-return
                    self._layer_colors[self._pointer] = next(self._log_colors)
                    change_color = False
            else:
                # Merge branch.
                change_color = True
                self._merge_branches(parent)
            yield self._graph_printer(
                current_node,
                original_pointer,
            )

            if not parent:
                break
            self._set_next_node(parent)


def _echo_messages(message_generator: Iterable[str]) -> None:
    if sys.platform.startswith("win"):
        for item in message_generator:
            click.echo(item)
    else:
        click.echo_via_pager(message_generator)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloaders of 5 categories AnimalPose dataset and 7 categories AnimalPose dataset."""

from tensorbay.opendataset.AnimalPose.loader import AnimalPose5, AnimalPose7

__all__ = ["AnimalPose5", "AnimalPose7"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the COVID_CT dataset."""

from tensorbay.opendataset.COVID_CT.loader import COVID_CT

__all__ = ["COVID_CT"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas tag."""

import click

from tensorbay.cli.tbrn import TBRN, TBRNType
from tensorbay.cli.utility import (
    ContextInfo,
    error,
    exception_handler,
    get_dataset_client,
    sort_branches_or_tags,
)
from tensorbay.client.gas import DatasetClientType


@exception_handler
def _implement_tag(obj: ContextInfo, tbrn: str, name: str, is_delete: bool, sort_key: str) -> None:
    tbrn_info = TBRN(tbrn=tbrn)

    if tbrn_info.type != TBRNType.DATASET:
        error(f'To operate a tag, "{tbrn}" must be a dataset')

    gas = obj.get_gas()
    dataset_client = get_dataset_client(gas, tbrn_info)

    if is_delete:
        _delete_tag(dataset_client, tbrn_info)
    elif name:
        _create_tag(dataset_client, name)
    else:
        _list_tags(dataset_client, sort_key)


def _delete_tag(dataset_client: DatasetClientType, tbrn_info: TBRN) -> None:
    if not tbrn_info.revision:
        error(f'To delete a tag, "{tbrn_info.get_tbrn()}" must have a tag name')

    dataset_client.delete_tag(tbrn_info.revision)
    tag_tbrn = TBRN(dataset_client.name, revision=tbrn_info.revision).get_colored_tbrn()
    click.echo(f'Successfully deleted tag "{tag_tbrn}"')


def _create_tag(dataset_client: DatasetClientType, name: str) -> None:
    if dataset_client.status.is_draft:
        error(f'To create a tag, "{dataset_client.name}" cannot be in draft status')

    if not dataset_client.status.commit_id:
        error(f'To create a tag, "{dataset_client.name}" should have commit history')

    dataset_client.create_tag(name=name)
    tag_tbrn = TBRN(dataset_client.name, revision=name).get_colored_tbrn()
    click.echo(f'Successfully created tag "{tag_tbrn}"')


def _list_tags(dataset_client: DatasetClientType, sort_key: str) -> None:
    for tag in sort_branches_or_tags(sort_key, dataset_client.list_tags()):
        click.echo(tag.name)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of FSDD dataset."""

import os

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification
from tensorbay.opendataset._utility import glob

DATASET_NAME = "FSDD"

_METADATA = {
    "jackson": {"gender": "male", "accent": "USA/neutral", "language": "english"},
    "nicolas": {"gender": "male", "accent": "BEL/French", "language": "english"},
    "theo": {"gender": "male", "accent": "USA/neutral", "language": "english"},
    "yweweler": {"gender": "male", "accent": "DEU/German", "language": "english"},
    "george": {"gender": "male", "accent": "GRC/Greek", "language": "english"},
    "lucas": {"gender": "male", "accent": "DEU/German", "language": "english"},
}


def FSDD(path: str) -> Dataset:
    """`Free Spoken Digit <https://github.com/Jakobovski/free-spoken-digit-dataset>`_ dataset.

    The file structure should be like::

        <path>
            recordings/
                0_george_0.wav
                0_george_1.wav
                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    label_map = {}
    for key, value in _METADATA.items():
        attributes = {"name": key}
        attributes.update(value)
        label_map[key] = attributes

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    segment = dataset.create_segment()
    audio_paths = glob(os.path.join(path, "recordings", "*.wav"))
    for audio_path in audio_paths:
        category, name = os.path.basename(audio_path).split("_")[:2]
        data = Data(audio_path)
        data.label.classification = Classification(category, label_map[name])
        segment.append(data)
    return dataset



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of LISATrafficSign dataset."""

import csv
import os

from tensorbay.dataset import Data, Dataset, Segment
from tensorbay.label import LabeledBox2D
from tensorbay.opendataset._utility import glob

DATASET_NAME = "LISATrafficSign"


def LISATrafficSign(path: str) -> Dataset:
    """`LISA Traffic Sign <https://git-disl.github.io/GTDLBench/datasets\
    /lisa_traffic_sign_dataset/>`_ dataset.

    The file structure should be like::

        <path>
            readme.txt
            allAnnotations.csv
            categories.txt
            datasetDescription.pdf
            videoSources.txt
            aiua120214-0/
                frameAnnotations-DataLog02142012_external_camera.avi_annotations/
                    diff.txt
                    frameAnnotations.bak
                    frameAnnotations.bak2
                    frameAnnotations.csv
                    keepRight_1330547092.avi_image10.png
                    keepRight_1330547092.avi_image11.png
                    keepRight_1330547092.avi_image12.png
                    ...
            aiua120214-1/
                frameAnnotations-DataLog02142012_001_external_camera.avi_annotations/
            aiua120214-2/
                frameAnnotations-DataLog02142012_002_external_camera.avi_annotations/
            aiua120306-0/
                frameAnnotations-DataLog02142012_002_external_camera.avi_annotations/
            aiua120306-1/
                frameAnnotations-DataLog02142012_003_external_camera.avi_annotations/
            vid0/
                frameAnnotations-vid_cmp2.avi_annotations/
            vid1/
                frameAnnotations-vid_cmp1.avi_annotations/
            vid10/
                frameAnnotations-MVI_0122.MOV_annotations/
            vid11/
                frameAnnotations-MVI_0123.MOV_annotations/
            vid2/
                frameAnnotations-vid_cmp2.avi_annotations/
            vid3/
                frameAnnotations-vid_cmp2.avi_annotations/
            vid4/
                frameAnnotations-vid_cmp2.avi_annotations/
            vid5/
                frameAnnotations-vid_cmp2.avi_annotations/
            vid6/
                frameAnnotations-MVI_0071.MOV_annotations/
            vid7/
                frameAnnotations-MVI_0119.MOV_annotations/
            vid8/
                frameAnnotations-MVI_0120.MOV_annotations/
            vid9/
                frameAnnotations-MVI_0121.MOV_annotations/
            negatives/
                negativePics/
                negatives.dat
            tools/
                evaluateDetections.py
                extractAnnotations.py
                mergeAnnotationFiles.py
                splitAnnotationFiles.py

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for dir_entry in os.scandir(root_path):
        if not dir_entry.is_dir() or not dir_entry.name.startswith(("vid", "aiua")):
            continue
        dataset.add_segment(_load_positive_segment(dir_entry.name, dir_entry.path))
    dataset.add_segment(_load_negative_segment(root_path))

    return dataset


def _load_positive_segment(segment_name: str, segment_path: str) -> Segment:
    if segment_name.startswith("vid"):
        # Pad zero for segment name to change "vid0" to "vid00"
        segment_name = f"{segment_name[:3]}{int(segment_name[3:]):02}"
    segment = Segment(segment_name)
    annotation_file = glob(
        os.path.join(segment_path, "frameAnnotations-*", "frameAnnotations.csv")
    )[0]
    image_folder = os.path.dirname(annotation_file)
    pre_filename = ""
    with open(annotation_file, "r", encoding="utf-8") as fp:
        for annotation in csv.DictReader(fp, delimiter=";"):
            filename = annotation["Filename"]

            if filename != pre_filename:
                data = Data(os.path.join(image_folder, filename))
                data.label.box2d = []
                segment.append(data)
                pre_filename = filename

            occluded, on_another_road = annotation["Occluded,On another road"].split(",", 1)
            data.label.box2d.append(
                LabeledBox2D(
                    int(annotation["Upper left corner X"]),
                    int(annotation["Upper left corner Y"]),
                    int(annotation["Lower right corner X"]),
                    int(annotation["Lower right corner Y"]),
                    category=annotation["Annotation tag"],
                    attributes={
                        "Occluded": bool(int(occluded)),
                        "On another road": bool(int(on_another_road)),
                        "Origin file": annotation["Origin file"],
                        "Origin frame number": int(annotation["Origin frame number"]),
                        "Origin track": annotation["Origin track"],
                        "Origin track frame number": int(annotation["Origin track frame number"]),
                    },
                )
            )
    return segment


def _load_negative_segment(root_path: str) -> Segment:
    segment = Segment("negative")
    for negative_image_path in glob(os.path.join(root_path, "negatives", "negativePics", "*.png")):
        data = Data(negative_image_path)
        data.label.box2d = []
        segment.append(data)
    return segment



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of Newsgroups20 dataset."""

import os

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification
from tensorbay.opendataset._utility import glob

DATASET_NAME = "Newsgroups20"
SEGMENT_DESCRIPTION_DICT = {
    "20_newsgroups": "Original 20 Newsgroups data set",
    "20news-bydate-train": (
        "Training set of the second version of 20 Newsgroups, "
        "which is sorted by date and has duplicates and some headers removed"
    ),
    "20news-bydate-test": (
        "Test set of the second version of 20 Newsgroups, "
        "which is sorted by date and has duplicates and some headers removed"
    ),
    "20news-18828": (
        "The third version of 20 Newsgroups, which has duplicates removed "
        "and includes only 'From' and 'Subject' headers"
    ),
}


def Newsgroups20(path: str) -> Dataset:
    """`20 Newsgroups <http://qwone.com/~jason/20Newsgroups/>`_ dataset.

    The folder structure should be like::

        <path>
            20news-18828/
                alt.atheism/
                    49960
                    51060
                    51119
                    51120
                    ...
                comp.graphics/
                comp.os.ms-windows.misc/
                comp.sys.ibm.pc.hardware/
                comp.sys.mac.hardware/
                comp.windows.x/
                misc.forsale/
                rec.autos/
                rec.motorcycles/
                rec.sport.baseball/
                rec.sport.hockey/
                sci.crypt/
                sci.electronics/
                sci.med/
                sci.space/
                soc.religion.christian/
                talk.politics.guns/
                talk.politics.mideast/
                talk.politics.misc/
                talk.religion.misc/
            20news-bydate-test/
            20news-bydate-train/
            20_newsgroups/

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for segment_name, segment_description in SEGMENT_DESCRIPTION_DICT.items():
        segment_path = os.path.join(root_path, segment_name)
        if not os.path.isdir(segment_path):
            continue

        segment = dataset.create_segment(segment_name)
        segment.description = segment_description

        text_paths = glob(os.path.join(segment_path, "*", "*"))
        for text_path in text_paths:
            category = os.path.basename(os.path.dirname(text_path))

            data = Data(
                text_path, target_remote_path=f"{category}/{os.path.basename(text_path)}.txt"
            )
            data.label.classification = Classification(category)
            segment.append(data)

    return dataset



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of Elpv dataset."""

import os

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification

DATASET_NAME = "Elpv"


def Elpv(path: str) -> Dataset:
    """`elpv <https://github.com/zae-bayern/elpv-dataset>`_ dataset.

    The file structure should be like::

        <path>
            labels.csv
            images/
                cell0001.png
                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    segment = dataset.create_segment()

    csv_path = os.path.join(root_path, "labels.csv")

    with open(csv_path, encoding="utf-8") as csv_file:
        for row in csv_file:
            image_name, attributes, category = row.strip().split()
            dirname, basename = image_name.split("/")
            image_path = os.path.join(root_path, dirname, basename)
            data = Data(image_path)
            data.label.classification = Classification(
                attributes={"defect probability": float(attributes)}, category=category
            )
            segment.append(data)
    return dataset
